"""lg-rez / blocs / Émulation de terminal Python (avancé)

Permet d'utiliser un terminal Python directement dans Discord
(implémentation de https://pypi.org/project/asyncode).

"""

import re
import sys
import enum
import asyncio

import asyncode
import discord

from lgrez import config
from lgrez.blocs import tools
from lgrez.blocs.journey import DiscordJourney


def _py_bloc(data):
    return tools.code_bloc(data, "py")


class RealShellExit(RuntimeError):
    """Terminal exited (:exc:`SystemExit` catched or stop button).

    Derivates from :exc:`RuntimeError`.
    """

    pass


class RSCommand(enum.Enum):
    """:class:`~enum.Enum`: Shell controll commands.

    These commands allow to control the shell in ways impossible with
    plain Discord messages: change indentation (leading spaces in
    messages are stripped), send empty lines, send EOF (^D).

    Enum values are unicode emojis codepoints used for emoji control
    of the shell. Override this class (you may not define all values to
    disable some features) to change them.

    Emojis are displayed in the same order members are defined:

    Attributes:
        HOME: Reset indentation to 0.
        UNTAB: Decreases indentation by one level.
        TAB: Increases indentation by one level.
        ENTER: Send empty line.
        EOF: Send EOF signal (^D in \*nix, ^Z + Enter in Windows.)
    """

    HOME = "\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}"
    UNTAB = "\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}"
    TAB = "\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}"
    ENTER = "\N{LEFTWARDS ARROW WITH HOOK}"  # close to "enter" sign
    EOF = "\N{BLACK SQUARE FOR STOP}"


class RealShell(asyncode.AsyncInteractiveConsole):
    """Python shell inside LGBot.

    (subclass of :class:`asyncode.AsyncInteractiveConsole`)

    Args:
        journey: Chan in which run the shell (show and wait for commands).
            While the shell is running, every message in this journey will
            be interpreted and deleted by this shell.
        locals, filename: passed to
            :class:`asyncode.AsyncInteractiveConsole`.
    """

    def __init__(
        self,
        journey: DiscordJourney,
        locals: dict = {},
        filename: str = "/shell",
    ) -> None:
        """Initialize self"""
        self.bot = config.bot
        self.journey = journey

        self.message: discord.Message = None
        self.content = ""
        self.indent = 0
        self.rsbuffer = ""

        locals["_shell"] = self
        super().__init__(locals, filename)

    async def interact(
        self,
        banner: str | None = None,
        exitmsg: str | None = "Byebye!",
    ):
        """Launches the shell.

        (see :meth:`asyncode.AsyncInteractiveConsole.interact`)
        """
        wait = _py_bloc("Launching Python shell...\n")
        *_, self.message = await self.journey.send(wait)

        if banner is None:
            base = (
                '# Type "help", "copyright", "credits" or "license" for more information.\n'
                '# async REPL: use "await" directly instead of asyncio.run().\n'
            )
            dsc = (
                f"# Discord implementation: the caret (‸) indicates the current position of the cursor.\n"
                '# Use control buttons or send "<", ">", "=" to (un)indent it and send empty lines.'
            )
            banner = f"# Python {sys.version} on {sys.platform}\n{base}\n{dsc}\n"

        try:
            await super().interact(banner, exitmsg)
        except SystemExit as exc:
            await self.write(f"\n{exitmsg}", refresh=True, show_caret=False)
            raise RealShellExit(*exc.args)
        finally:
            await self.message.clear_reactions()

    async def control(self, command: RSCommand, times: int = 1) -> str | None:
        """Applies the effect of a shell control.

        Args:
            command: shell command to compute.
            times: repeat the command several times. Makes ony sense
                for :attr:`.RSCommand.TAB` / :attr:`.RSCommand.UNTAB`.

        Returns:
            ``None`` if further instructions are needed, else the
            complete instruction to push into the shell.

        Raises:
            .RealShellExit: if asked to stop shell (EOF).
        """
        if not isinstance(command, RSCommand):
            command = RSCommand(command)

        if command is RSCommand.HOME:  # Reset indentation
            if self.indent > 0:
                self.indent = 0
                await self.write("", refresh=True)
        if command is RSCommand.UNTAB:  # Unindent
            if self.indent > 0:
                self.indent -= min(times, self.indent)
                await self.write("", refresh=True)
        elif command is RSCommand.TAB:  # Indent
            self.indent += times
            await self.write("", refresh=True)
        elif command is RSCommand.ENTER:  # Newline
            # Send empty str (impossible through Discord)
            return ""
        elif command is RSCommand.EOF:  # STOP
            raise RealShellExit("Arrêt du shell.")

        return None

    async def _check_reacts(self) -> bool | tuple[bool, str | None]:
        """Preliminary step of self.get_order().

        Returns (True, <value to be returned>) if a react has been
        anticipated, else False.
        """
        reacts = self.message.reactions  # Réactions déjà présentes
        if not reacts:
            # On ajoute les emojis si pas déjà présents (nouveau message)
            for command in RSCommand:
                await self.message.add_reaction(command.value)
        else:
            # Sinon, on regarde si un emoji est déjà cliqué
            try:
                react = next(r for r in reacts if r.count > 1)
            except StopIteration:
                # Pas le cas ==> on poursuit
                pass
            else:
                # C'est le cas : on trouve l'utilisateur
                async for user in react.users():
                    if user != self.bot.user:
                        # On enlève la réaction
                        await react.remove(user)
                        break
                command = RSCommand(react.emoji.name)
                # Et on applique la commande
                control_ret = await self.control(command)
                return (True, control_ret)

        return False

    async def get_order(self) -> str | None:
        """Waits for an order (message or reaction) from a member.

        Add reactions if not present (new message) and waits for
        an interaction.

        Returns:
            ``None`` if further instructions are needed, else the
            complete instruction to push into the shell.

        Raises:
            .RealShellExit: if asked to stop shell (EOF).
        """
        res = await self._check_reacts()
        if res:  # Réaction anticipée, control effectué
            return res[1]

        def react_check(payload: discord.RawReactionActionEvent):
            # Check REACT : bon message et pas react du bot
            return payload.message_id == self.message.id and payload.user_id != self.bot.user.id

        def message_check(mess: discord.Message):
            # Check MESSAGE : bon channel et pas du bot
            return mess.channel == self.journey.channel and mess.author != self.bot.user

        react_task = asyncio.create_task(self.bot.wait_for("raw_reaction_add", check=react_check))
        mess_task = asyncio.create_task(tools.wait_for_message(check=message_check, chan=self.journey.channel))
        # On exécute les deux tâche concurremment
        done, pending = await asyncio.wait({react_task, mess_task}, return_when=asyncio.FIRST_COMPLETED)

        for task in pending:
            task.cancel()
        done_task = next(iter(done))  # done = tâche complétée

        if done_task == react_task:  # Clic sur emoji de contrôle
            payload = done_task.result()
            # On enlève la réaction
            await self.message.remove_reaction(payload.emoji, payload.member)
            command = RSCommand(payload.emoji.name)
            return await self.control(command)

        else:  # Réponse par message / STOP
            message = done_task.result()
            text = message.content
            await message.delete()

            if match := re.match("^<+", text):
                await self.control(RSCommand.UNTAB, times=match.end())
                text = text[match.end() :]
            elif match := re.match("^>+", text):
                await self.control(RSCommand.TAB, times=match.end())
                text = text[match.end() :]

            if text == "=":
                return await self.control(RSCommand.ENTER)
            else:
                return text or None

    async def write(
        self,
        data: str,
        *,
        refresh: bool = False,
        show_caret: bool = True,
    ) -> None:
        """Method called on each print / repr / traceback...

        Updates Discord message only if text ends with a prompt
        (:attr:`sys.ps1` or :attr:`sys.ps2`); adds it to an internal
        buffer otherwise.

        Args:
            data: text to display.
            refresh: if ``True``, updates message whatever ``data`` is.
            show_caret: if ``False``, does not display the cursor
                position (‸); implies that the shell is not ready for
                further instructions.
        """
        if not hasattr(sys, "ps1"):
            sys.ps1 = ">>> "
        if not hasattr(sys, "ps2"):
            sys.ps2 = "... "

        if refresh or data.endswith(sys.ps1) or data.endswith(sys.ps2):
            # Prompt / refresh demandé
            data = self.rsbuffer + data  # On vide le buffer
            self.rsbuffer = ""
        else:
            self.rsbuffer += data  # On remplit le buffer
            return

        # Auto-indent à l'entrée dans un nouveau bloc
        if data.endswith(f":\n{sys.ps2}"):
            self.indent += 1

        # Ajout caret d'insertion pour indiquer la position du curseur
        if show_caret:
            insert = " " * (self.indent * 4) + "‸"
        else:
            insert = ""

        # Actualisation message / nouvau si plein
        if len(self.content + data + insert) < 1990:
            self.content += data
            await self.message.edit(content=_py_bloc(self.content + insert))
        else:
            # On enlève les commandes du message
            await self.message.clear_reactions()

            blocs = tools.smooth_split(data, N=1990)
            # On modifie le texte actuel
            self.content = blocs[-1]
            blocs[-1] += insert

            for bloc in blocs:
                message = await self.journey.send(_py_bloc(bloc))

            # On modifie le message actuel = dernier envoyé
            self.message = message

    async def raw_input(self, prompt: str = "") -> str:
        """Method called when the shell waits for next instruction.

        Calls :meth:`.get_order` until it returns a :class:`str` (or
        propagates :exc:`.RealShellExit`) then displays it in the shell
        (without caret) before returning its value.

        Args:
            prompt: optional text to send before waiting for input.

        Returns:
            The new instruction to compute.
        """
        if prompt:
            await self.write(prompt, refresh=True)

        entree = None
        while entree is None:
            # Modifie l'indentation, stop ou récupère une ligne
            entree = await self.get_order()

        entree = " " * (self.indent * 4) + entree
        # Affiche l'entrée
        await self.write(entree + "\n", refresh=True, show_caret=False)
        return entree



"""lg-rez / features / Tâches planifiées

Planification, liste, annulation, exécution de tâches planifiées

"""

import datetime

import discord
from discord import app_commands

from lgrez import config, commons
from lgrez.blocs import env, gsheets, tools
from lgrez.bdd import (
    Joueur,
    Action,
    Role,
    Camp,
    Utilisation,
    Ciblage,
    BaseCiblage,
    CandidHaro,
    CandidHaroType,
    UtilEtat,
    CibleType,
    Vote,
)
from lgrez.blocs.journey import DiscordJourney, journey_command
from lgrez.features import gestion_actions


async def export_vote(vote: Vote | None, utilisation: Utilisation) -> None:
    """Enregistre un vote/les actions résolues dans le GSheet ad hoc.

    Écrit dans le GSheet ``LGREZ_DATA_SHEET_ID``. Peut être écrasé
    pour une autre implémentation.

    Args:
        vote: le vote concerné, ou ``None`` pour une action.
        utilisation: l'utilisation qui vient d'être effectuée.
            Doit être remplie (:attr:`.bdd.Utilisation.is_filled`).

    Raises:
        RuntimeError: si la variable d'environnement ``LGREZ_DATA_SHEET_ID``
            n'est pas définie.

    Note:
        Fonction asynchrone depuis la version 2.2.2.
    """
    if vote and not isinstance(vote, Vote):
        vote = Vote[vote]  # str -> Vote

    joueur = utilisation.action.joueur
    match vote:
        case Vote.cond:
            sheet_name = config.db_votecond_sheet
            data = [joueur.nom, utilisation.cible.nom]
        case Vote.maire:
            sheet_name = config.db_votemaire_sheet
            data = [joueur.nom, utilisation.cible.nom]
        case Vote.loups:
            sheet_name = config.db_voteloups_sheet
            data = [joueur.nom, joueur.camp.slug, utilisation.cible.nom]
        case _:
            sheet_name = config.db_actions_sheet
            recap = "\n+\n".join(
                f"{action.base.slug}({last_util.decision})"
                for action in joueur.actions_actives
                if (
                    (last_util := action.derniere_utilisation)
                    and last_util.is_filled  # action effectuée
                    and last_util.ts_decision.date() == datetime.date.today()
                )
            )
            data = [joueur.nom, joueur.role.slug, joueur.camp.slug, recap]

    LGREZ_DATA_SHEET_ID = env.load("LGREZ_DATA_SHEET_ID")
    workbook = await gsheets.connect(LGREZ_DATA_SHEET_ID)
    sheet = await workbook.worksheet(sheet_name)
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    await sheet.append_row([timestamp, *data], value_input_option="USER_ENTERED")


async def check_last_utilisation(
    action: Action, base_ciblage: BaseCiblage, cible: Joueur | Role | Camp | bool | str
) -> None:
    """Demande une cible à l'utilisateur.

    Args:
        ctx: le contexte de commande.
        action: action pour laquelle on cherche une cible.
        base_ciblage: ciblage à demander.
        first: proposition initiale du joueur (passée comme argument d'une commande).

    Returns:
        La cible sélectionnée, selon le type de ciblage.

    Réalise les interactions adéquates en fonction du type du base_ciblage,
    vérifie le changement de cible le cas échéant.
    """
    if not base_ciblage.doit_changer:
        return

    derniere_util = action.utilisations.filter(~Utilisation.is_open).order_by(Utilisation.ts_close.desc()).first()
    if derniere_util and derniere_util.etat == UtilEtat.validee:
        # Dernière utilisation validée : comparaison avec ciblages
        # de même prio que le ciblage en cours de demande
        cibles = [cib.valeur for cib in derniere_util.ciblages if cib.base.prio == base_ciblage.prio]

        if cible in cibles:  # interdit !
            raise commons.UserInputError(
                "cible",
                f":stop_sign: {cible} déjà ciblé(e) lors de la précédente utilisation, merci de changer :stop_sign:\n"
                "*(`@MJ` si contestation)*",
            )


DESCRIPTION = """Commandes de vote et d'action de rôle"""


async def do_vote(journey: DiscordJourney, vote: Vote, votant: Joueur, cible: Joueur):
    match vote:
        case Vote.cond:
            vote_name = "le condamné du jour"
            pour_contre = "contre"
            can_vote = votant.votant_village
        case Vote.maire:
            vote_name = "le nouveau maire"
            pour_contre = "pour"
            can_vote = votant.votant_village
        case Vote.loups:
            vote_name = "la victime du soir"
            pour_contre = "contre"
            can_vote = votant.votant_loups

    try:
        vaction = votant.action_vote(vote)
    except RuntimeError:
        await journey.send(":x: Minute papillon, le jeu n'est pas encore lancé !")
        return

    if not can_vote:
        await journey.send(":x: Tu n'as pas le droit de participer à ce vote.")
        return
    if not vaction.is_open:
        await journey.send(f":x: Pas de vote pour {vote_name} en cours !")
        return

    util = vaction.derniere_utilisation

    # Test si la cible est sous le coup d'un haro / candidate
    if vote == Vote.cond and not CandidHaro.query.filter_by(joueur=cible, type=CandidHaroType.haro).first():
        await journey.ok_cancel(
            f"{cible.nom} n'a pas (encore) subi ou posté de haro ! "
            "Si c'est toujours le cas à la fin du vote, ton vote sera compté comme blanc... \n"
            "Veux-tu continuer ?"
        )
    elif vote == Vote.maire and not CandidHaro.query.filter_by(joueur=cible, type=CandidHaroType.candidature).first():
        await journey.ok_cancel(
            f"{cible.nom} ne s'est pas (encore) présenté(e) ! "
            "Si c'est toujours le cas à la fin de l'élection, ton vote sera compté comme blanc... \n"
            "Veux-tu continuer ?"
        )

    if not vaction.is_open:
        # On revérifie, si ça a fermé entre temps !!
        await journey.send(f":x: Le vote pour {vote_name} a fermé entre temps, pas de chance !")
        return

    # Modification en base
    if util.ciblages:  # ancien ciblage
        Ciblage.delete(*util.ciblages)
    Ciblage(utilisation=util, joueur=cible).add()
    util.ts_decision = datetime.datetime.now()
    util.etat = UtilEtat.remplie
    util.update()

    # Écriture dans sheet Données brutes
    await export_vote(vote, util)

    confirm = f"Vote {pour_contre} {tools.bold(cible.nom)} bien pris en compte."
    [message] = await journey.send(
        f"{confirm}\n" + tools.ital("Tu peux modifier ton vote autant que nécessaire avant sa fermeture.")
    )
    if message.channel != votant.private_chan:
        await votant.private_chan.send(confirm)


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def vote(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.HaroteTransformer]):
    """Vote pour le condamné du jour.

    Args:
        joueur: Le joueur contre qui tu veux diriger ton vote.

    Cette commande n'est utilisable que lorsqu'un vote pour le condamné est en cours,
    pour les joueurs ayant le droit de voter.

    Le bot t'enverra un message à l'ouverture de chaque vote.

    La commande peut être utilisée autant que voulu pour changer de cible tant que le vote est en cours.
    """
    moi = Joueur.from_member(journey.member)
    await do_vote(journey, Vote.cond, votant=moi, cible=joueur)


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def votemaire(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.CandidatTransformer]):
    """Vote pour le nouveau maire.

    Args:
        joueur: Le joueur pour lequel tu souhaites voter.

    Cette commande n'est utilisable que lorsqu'une élection pour le maire est en cours,
    pour les joueurs ayant le droit de voter.

    Le bot t'enverra un message à l'ouverture de chaque vote.

    La commande peut être utilisée autant que voulu pour changer de cible tant que le vote est en cours.
    """
    moi = Joueur.from_member(journey.member)
    await do_vote(journey, Vote.maire, votant=moi, cible=joueur)


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def voteloups(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.VivantTransformer]):
    """Vote pour la victime de l'attaque des loups (si tu es un loup, évidemment...)

    Args:
        joueur: Le joueur que tu souhaites éliminer.

    Cette commande n'est utilisable que lorsqu'une vote pour la victime du soir est en cours,
    pour les joueurs concernés.

    Le bot t'enverra un message à l'ouverture de chaque vote.

    La commande peut être utilisée autant que voulu pour changer de cible tant que le vote est en cours.
    """
    moi = Joueur.from_member(journey.member)
    await do_vote(journey, Vote.loups, votant=moi, cible=joueur)


class ActionTransformer(app_commands.Transformer):
    async def transform(self, interaction: discord.Interaction, value: str) -> Action:
        return Action.query.get(int(value))

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        async with DiscordJourney(interaction) as journey:
            try:
                joueur = Joueur.from_member(journey.member)
            except ValueError:
                return []
            return [
                app_commands.Choice(name=action.base.slug, value=str(action.id))
                for action in joueur.actions_actives
                if action.is_open
            ][:25]


class CibleTransformer(app_commands.Transformer):
    N_CIBLE = 0

    ERROR_CHOICE = app_commands.Choice(name="⚠️ Remplir d'abord le paramètre 'action' ⚠️", value="error")
    VALIDATE_CHOICE = app_commands.Choice(name="Valider", value="__NO_CIBLE")
    YES_CHOICE = app_commands.Choice(name="Oui", value="yes")
    NO_CHOICE = app_commands.Choice(name="Non", value="no")

    PHRASE_CHOICE_NAME_PREFIX = "🔽"
    PHRASE_CHOICE_VALUE = "__PHRASE"

    async def transform(self, interaction: discord.Interaction, value: str) -> Joueur | Role | Camp | bool | str | None:
        if value in (self.VALIDATE_CHOICE.name, self.VALIDATE_CHOICE.value):
            return None
        if value in (self.ERROR_CHOICE.name, self.ERROR_CHOICE.value):
            raise commons.UserInputError(
                f"cible_{self.N_CIBLE}" if self.N_CIBLE else "cible",
                "Paramètre `action` non rempli !",
            )
        if value.startswith(self.PHRASE_CHOICE_NAME_PREFIX) or value == self.PHRASE_CHOICE_VALUE:
            raise commons.UserInputError(
                f"cible_{self.N_CIBLE}" if self.N_CIBLE else "cible",
                "L'en-tête d'explication des choix n'est pas un choix valide !",
            )

        action_id = interaction.namespace.action
        action: Action = Action.query.get(int(action_id))
        base_ciblage = action.base.base_ciblages[self.N_CIBLE]

        match base_ciblage.type:
            case CibleType.joueur:
                return await tools.JoueurTransformer().transform(interaction, value)
            case CibleType.vivant:
                return await tools.VivantTransformer().transform(interaction, value)
            case CibleType.mort:
                return await tools.MortTransformer().transform(interaction, value)
            case CibleType.role:
                return await tools.RoleTransformer().transform(interaction, value)
            case CibleType.camp:
                return await tools.CampTransformer().transform(interaction, value)
            case CibleType.booleen:
                return value in (self.YES_CHOICE.name, self.YES_CHOICE.value)
            case CibleType.texte:
                return value

    async def autocomplete(self, interaction: discord.Interaction, current: str) -> list[app_commands.Choice[str]]:
        action_id = interaction.namespace.action
        if not action_id:
            return [self.ERROR_CHOICE]

        action: Action = Action.query.get(int(action_id))
        try:
            base_ciblage = action.base.base_ciblages[self.N_CIBLE]
        except IndexError:
            # Pas autant de ciblages pour cette action
            return [self.VALIDATE_CHOICE]

        match base_ciblage.type:
            case CibleType.joueur:
                choices = await tools.JoueurTransformer().autocomplete(interaction, current)
            case CibleType.vivant:
                choices = await tools.VivantTransformer().autocomplete(interaction, current)
            case CibleType.mort:
                choices = await tools.MortTransformer().autocomplete(interaction, current)
            case CibleType.role:
                choices = await tools.RoleTransformer().autocomplete(interaction, current)
            case CibleType.camp:
                choices = await tools.CampTransformer().autocomplete(interaction, current)
            case CibleType.booleen:
                choices = [self.YES_CHOICE, self.NO_CHOICE]
            case CibleType.texte:
                choices = [app_commands.Choice(name="[Texte libre]", value=current)]

        return [
            app_commands.Choice(
                name=f"{self.PHRASE_CHOICE_NAME_PREFIX}  {base_ciblage.phrase} 🔽"[:100], value=self.PHRASE_CHOICE_VALUE
            ),
            *choices[:24],
        ]


class Cible2Transformer(CibleTransformer):
    N_CIBLE = 1


class Cible3Transformer(CibleTransformer):
    N_CIBLE = 2


@app_commands.command(name="action")
@tools.vivants_only
@tools.private()
@journey_command
async def action_(
    journey: DiscordJourney,
    *,
    action: app_commands.Transform[Action, ActionTransformer],
    cible: app_commands.Transform[Joueur | Role | Camp | bool | str | None, CibleTransformer],
    cible_2: app_commands.Transform[Joueur | Role | Camp | bool | str | None, Cible2Transformer],
    cible_3: app_commands.Transform[Joueur | Role | Camp | bool | str | None, Cible3Transformer] | None = None,
):
    """Utilise l'action de ton rôle / une des actions associées.

    Args:
        action: L'action pour laquelle agir (si il n'y a pas de suggestions, c'est que tu ne peux pas agir !)
        cible: La première cible de l'action. Voir la première suggestion pour plus de détails
        cible_2: La deuxième cible de l'action, si nécessaire. Voir la première suggestion pour plus de détails
        cible_3: La troisième cible de l'action, si nécessaire. Voir la première suggestion pour plus de détails

    Cette commande n'est utilisable que si tu as au moins une action ouverte.
    Action = pouvoir associé à ton rôle, mais aussi pouvoirs ponctuels (Lame Vorpale, Chat d'argent...)
    Le bot t'enverra un message à l'ouverture de chaque action.

    La commande peut être utilisée autant que voulu pour changer d'action tant que la fenêtre d'action est en cours,
    SAUF pour certaines actions (dites "instantanées") ayant une conséquence immédiate (Barbier, Licorne...).
    Le bot mettra dans ce cas un message d'avertissement.
    """
    joueur = Joueur.from_member(journey.member)

    # Vérification rôle actif
    if not joueur.role_actif:
        await journey.send(":x: Tu ne peux pas utiliser tes pouvoirs pour le moment !")
        return

    # Détermine la/les actions en cours pour le joueur
    actions = [ac for ac in joueur.actions_actives if ac.is_open]
    if not actions:
        await journey.send(":x: Aucune action en cours pour toi.")
        return

    util = action.derniere_utilisation

    if not cible:
        if util.ciblages:
            await journey.ok_cancel(f"Action actuelle : {tools.bold(action.decision)}\n\nAnnuler l'action ?")
            # Annulation de l'action
            Ciblage.delete(*util.ciblages)
            util.ts_decision = datetime.datetime.now()
            util.etat = UtilEtat.ignoree
            util.update()
            await journey.send("Utilisation de l'action annulée.")
        else:
            await journey.send('Action non utilisée pour le moment. Remplir le paramètre "cible" pour agir !')
        return

    # Vérification nombre de cibles
    cibles = {}
    base_ciblages = action.base.base_ciblages
    await check_last_utilisation(action, base_ciblages[0], cible)
    cibles[base_ciblages[0]] = cible

    if len(base_ciblages) > 1:
        if not cible_2:
            raise commons.UserInputError("cible_2", "Cette action a besoin d'un second paramètre !")
        await check_last_utilisation(action, base_ciblages[1], cible_2)
        cibles[base_ciblages[1]] = cible_2

    if len(base_ciblages) > 2:
        if not cible_3:
            raise commons.UserInputError("cible_3", "Cette action a besoin d'un troisième paramètre !")
        await check_last_utilisation(action, base_ciblages[2], cible_3)
        cibles[base_ciblages[2]] = cible_3

    if not action.is_open:
        # On revérifie, si ça a fermé entre temps !!
        await journey.send("L'action a fermé entre temps, pas de chance !")
        return

    # Avertissement si action a conséquence instantanée (barbier...)
    if action.base.instant:
        await journey.ok_cancel(
            "Attention : cette action a une conséquence instantanée ! "
            "Si tu valides, tu ne pourras pas revenir en arrière.\n"
            "Ça part ?"
        )

    # Modification en base
    if util.ciblages:  # ancien ciblages
        Ciblage.delete(*util.ciblages)
    for bc, cible in cibles.items():
        cib = Ciblage(utilisation=util, base=bc)
        cib.valeur = cible  # affecte le bon attribut selon le bc.type

    util.ts_decision = datetime.datetime.now()
    util.etat = UtilEtat.remplie
    util.update()

    # Écriture dans sheet Données brutes
    await export_vote(None, util)

    # Conséquences si action instantanée
    if action.base.instant:
        await gestion_actions.close_action(action)
        await journey.send(tools.ital(f"[Allô {config.Role.mj.mention}, conséquence instantanée ici !]"))

    else:
        await journey.send(
            f"Action « {tools.bold(action.decision)} » bien prise en compte pour {tools.code(action.base.slug)}.\n"
            + tools.ital("Tu peux modifier ta décision autant que nécessaire avant la fin du créneau.")
        )



"""lg-rez / features / Gestion des votes et actions

Ouverture / fermeture / rappels des votes et actions (+ refill)

"""

import datetime
import enum
from typing import Literal

from discord import app_commands

from lgrez import commons, config
from lgrez.blocs import tools
from lgrez.blocs.journey import DiscordJourney, journey_command
from lgrez.features import gestion_actions, communication
from lgrez.bdd import (
    Joueur,
    Action,
    BaseAction,
    Tache,
    CandidHaro,
    Utilisation,
    CandidHaroType,
    ActionTrigger,
    Vote,
)
from lgrez.features.taches import planif_command


async def _get_joueurs(quoi: Literal["open", "close", "remind"], qui: Vote) -> list[Joueur]:
    """Récupère les joueurs concernés par la tâche /quoi <qui> [heure].

    Args:
        quoi: évènement, ``"open" / "close" / "remind"``.
        qui:
            ===========     ===========
            ``Vote``        pour le vote correspondant
            ``actions``     pour les actions commençant à ``heure``
            ``{id}``        pour une action précise (:attr:`bdd.Action.id`)
            ===========     ===========

        heure: si ``qui == "actions"``, heure associée (au format ``HHhMM``).

    Returns:
        La liste des joueurs concernés.
    """
    # Critère principal : présence/absence d'une action actuellement ouverte (et non traitée pour remind)
    criteres = {
        "open": ~Joueur.actions.any(Action.is_open, vote=qui),
        "close": Joueur.actions.any(Action.is_open, vote=qui),
        "remind": Joueur.actions.any(Action.is_waiting, vote=qui),
    }
    critere = criteres[quoi]
    if quoi == "open":
        # Open : le joueur doit en plus avoir votant_village/loups True
        if qui == Vote.loups:
            critere &= Joueur.votant_loups.is_(True)
        else:
            critere &= Joueur.votant_village.is_(True)

    return Joueur.query.filter(critere).all()


async def _get_actions(quoi: Literal["open", "close", "remind"], heure: str) -> list[Action]:
    # Si l'heure est précisée, on convertit "HHhMM" -> datetime.time
    tps = tools.heure_to_time(heure)
    return gestion_actions.get_actions(quoi, ActionTrigger.temporel, tps)


async def _get_action(quoi: Literal["open", "close", "remind"], id: int) -> Action | None:
    action = Action.query.get(id)
    if not action:
        raise commons.UserInputError("qui", f"Pas d'action d'ID = {id}")
    if not action.active:
        raise commons.UserInputError("qui", f"Action d'ID = {id} inactive")

    # Appel direct action par son numéro (perma : rappel seulement)
    if (
        (quoi == "open" and (not action.is_open or action.base.trigger_debut == ActionTrigger.perma))
        or (quoi == "close" and action.is_open)
        or (quoi == "remind" and action.is_waiting)
    ):
        # Action lançable
        return action
    else:
        return None


async def _do_refill(motif: str, action: Action) -> None:
    # Détermination nouveau nombre de charges
    if motif in config.refills_full:
        # Refill -> nombre de charges initial de l'action
        new_charges = action.base.base_charges
    else:
        # Refill -> + 1 charge
        new_charges = action.charges + 1

    # Refill proprement dit
    if new_charges <= action.charges:
        # Pas de rechargement à faire (déjà base_charges)
        return

    if not action.charges and action.base.trigger_debut == ActionTrigger.perma:
        # Action permanente qui était épuisée : on ré-ouvre !
        if tools.en_pause():
            ts = tools.fin_pause()
        else:
            ts = datetime.datetime.now() + datetime.timedelta(seconds=10)
            # + 10 secondes pour ouvrir après le message de refill
        await planif_command(ts, open_action, id=action.id)

    action.charges = new_charges
    config.session.commit()

    await action.joueur.private_chan.send(
        f"Ton action {action.base.slug} vient d'être rechargée, "
        f"tu as maintenant {new_charges} charge(s) disponible(s) !"
    )


DESCRIPTION = """Commandes de gestion des votes et actions"""


open = app_commands.Group(name="open", description="Ouvrir quelque chose")
open = tools.mjs_only(open)


@open.command(name="vote")
@tools.mjs_only
@journey_command
async def open_vote(journey: DiscordJourney, *, qui: Vote, heure: str | None = None, heure_chain: str | None = None):
    """Lance un vote (COMMANDE BOT / MJ)

    Args:
        qui: Type de vote à lancer (vote pour le condamné du jour, le nouveau maire ou la victime des loups).
        heure: Heure à laquelle programmer la fermeture du vote, optionnel (HHh / HHhMM).
        heure_chain: Heure à laquelle programmer une ré-ouverture, pour boucler à l'infini (HHh / HHhMM).

    Une sécurité empêche de lancer un vote ou une action déjà en cours.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi
    (penser à envoyer / planifier la fermeture des votes, par exemple).

    Examples:
        - ``/open maire`` :         lance un vote maire maintenant
        - ``/open cond 19h`` :      lance un vote condamné maintenant et programme sa fermeture à 19h00
        - ``/open cond 18h 10h`` :  lance un vote condamné maintenant, programme sa fermeture à 18h00,
                                    et une prochaine ouverture à 10h qui se fermera à 18h, et ainsi de suite
    """
    joueurs = await _get_joueurs("open", qui)

    await journey.send(
        "\n".join(f" - {joueur.nom}" for joueur in joueurs),
        code=True,
        prefix=f"Joueur(s) répondant aux critères ({len(joueurs)}) :",
    )

    match qui:
        case Vote.cond:
            content = (
                f"{tools.montre()}  Le vote pour le condamné du jour est ouvert !  {config.Emoji.bucher} \n"
                + (f"Tu as jusqu'à {heure} pour voter. \n" if heure else "")
                + tools.ital(f"Tape {tools.code('/vote <joueur>')} pour voter.")
            )
            vote_command = "vote"
            haro_command = "haro"

        case Vote.maire:
            content = (
                f"{tools.montre()}  Le vote pour l'élection du maire est ouvert !  {config.Emoji.maire} \n"
                + (f"Tu as jusqu'à {heure} pour voter. \n" if heure else "")
                + tools.ital(f"Tape {tools.code('/votemaire <joueur>')} pour voter.")
            )
            vote_command = "votemaire"
            haro_command = "candid"

        case Vote.loups:
            content = (
                f"{tools.montre()}  Le vote pour la victime de cette nuit est ouvert !  {config.Emoji.lune} \n"
                + (f"Tu as jusqu'à {heure} pour voter. \n" if heure else "")
                + tools.ital(f"Tape {tools.code('/voteloups <joueur>')} pour voter.")
            )
            vote_command = "voteloups"
            haro_command = None

    # Activation commande de vote
    # config.bot.tree.enable_command(vote_command)
    # if haro_command:
    #     config.bot.tree.enable_command(haro_command)
    # await config.bot.tree.sync(guild=config.guild)

    # Création utilisations & envoi messages
    for joueur in joueurs:
        chan = joueur.private_chan

        action = joueur.action_vote(qui)
        if action.is_open:  # Sécurité : action ouverte depuis
            continue
        util = Utilisation(action=action)
        util.add()
        util.open()

        await chan.send(content)

    config.session.commit()

    # Actions déclenchées par ouverture
    if isinstance(qui, Vote):
        for action in Action.query.join(Action.base).filter(BaseAction.trigger_debut == ActionTrigger.open(qui)):
            await gestion_actions.open_action(action)

        for action in Action.query.join(Action.base).filter(BaseAction.trigger_fin == ActionTrigger.open(qui)):
            await gestion_actions.close_action(action)

    # Réinitialise haros/candids
    items = []
    if qui == Vote.cond:
        items = CandidHaro.query.filter_by(type=CandidHaroType.haro).all()
    elif qui == Vote.maire:
        items = CandidHaro.query.filter_by(type=CandidHaroType.candidature).all()
    if items:
        for item in items:
            await item.disable_message_buttons()
        CandidHaro.delete(*items)

        await tools.log(f"/open {qui.name} : haros/candids wiped")
        await config.Channel.haros.send(
            f"{config.Emoji.void}\n" * 30
            + "Nouveau vote, nouveaux haros !\n"
            + tools.ital(
                f"Les posts ci-dessus sont invalides pour le vote actuel. "
                f"Utilisez {tools.code('/haro')} pour en relancer."
            )
        )

    # Programme fermeture
    if heure:
        ts = tools.next_occurrence(tools.heure_to_time(heure))
        await planif_command(ts - datetime.timedelta(minutes=30), remind_vote, qui=qui)
        if heure_chain:
            await planif_command(ts, close_vote, qui=qui, heure=heure_chain, heure_chain=heure)
        else:
            await planif_command(ts, close_vote, qui=qui)


@open.command(name="actions")
@tools.mjs_only
@journey_command
async def open_actions(journey: DiscordJourney, heure: str):
    """Ouvre les actions commençant à une heure donnée (COMMANDE BOT / MJ)

    Args:
        heure: Heure de début des actions à lancer.

    Une sécurité empêche d'ouvrir une action déjà en cours.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi (penser à envoyer /
    planifier la fermeture de l'action).
    """
    actions = await _get_actions("open", heure)

    await journey.send(
        "\n".join(f" - {action.base.slug} - {action.joueur.nom}" for action in actions),
        code=True,
        prefix=f"Action(s) répondant aux critères ({len(actions)}) :",
    )
    for action in actions:
        await gestion_actions.open_action(action)


@open.command(name="action")
@tools.mjs_only
@journey_command
async def open_action(journey: DiscordJourney, id: int):
    """Lance une action donnée (COMMANDE BOT / MJ)

    Args:
        id: ID de l'action à ouvrir.

    Une sécurité empêche d'ouvrir une action déjà en cours.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi (penser à envoyer /
    planifier la fermeture de l'action).
    """
    action = await _get_action("open", id)
    if not action:
        await journey.send(f"L'action #{id} est déjà ouverte !")
        return

    await journey.send(f"Joueur concerné : {action.joueur}")
    await gestion_actions.open_action(action)


close = app_commands.Group(name="close", description="Clôturer quelque chose")
close = tools.mjs_only(close)


@close.command(name="vote")
@tools.mjs_only
@journey_command
async def close_vote(journey: DiscordJourney, *, qui: Vote, heure: str | None = None, heure_chain: str | None = None):
    """Ferme un vote (COMMANDE BOT / MJ)

    Args:
        qui: Type de vote à fermer (vote pour le condamné du jour, le nouveau maire ou la victime des loups).
        heure: Heure à laquelle programmer une prochaine ouverture du vote, optionnel (HHh / HHhMM).
        heure_chain: Heure à laquelle programmer une re-fermeture, pour boucler à l'infini (HHh / HHhMM).

    Une sécurité empêche de fermer un vote ou une action qui n'est pas en cours.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi
    (penser à envoyer / planifier la fermeture des votes, par exemple).

    Examples:
        - ``/close maire`` :        ferme le vote condamné maintenant
        - ``/close cond 10h`` :     ferme le vote condamné maintenant et programme une prochaine ouverture à 10h00
        - ``/close cond 10h 18h`` : ferme le vote condamné maintenant, programme une prochaine ouverture à 10h00,
                                    qui sera fermé à 18h, puis une nouvelle ouverture à 10h, etc
    """
    joueurs = await _get_joueurs("close", qui)

    await journey.send(
        "\n".join(f" - {joueur.nom}" for joueur in joueurs),
        code=True,
        prefix=f"Joueur(s) répondant aux critères ({len(joueurs)}) :",
    )

    match qui:
        case Vote.cond:
            content = (
                f"{tools.montre()}  Fin du vote pour le condamné du jour !\nVote définitif : {{nom_cible}}\n"
                f"Les résultats arrivent dans l'heure !\n"
            )
            vote_command = "vote"
            haro_command = "haro"

        case Vote.maire:
            content = f"{tools.montre()}  Fin du vote pour le maire ! \nVote définitif : {{nom_cible}}"
            vote_command = "votemaire"
            haro_command = "candid"

        case Vote.loups:
            content = f"{tools.montre()}  Fin du vote pour la victime du soir !\nVote définitif : {{nom_cible}}"
            vote_command = "voteloups"
            haro_command = None

    # Activation commande de vote
    # config.bot.tree.disable_command(vote_command)
    # if haro_command:
    #     config.bot.tree.disable_command(haro_command)
    # await config.bot.tree.sync(guild=config.guild)

    # Fermeture utilisations et envoi messages
    for joueur in joueurs:
        chan = joueur.private_chan

        if isinstance(qui, Vote):
            action = joueur.action_vote(qui)
            if not action.is_open:  # Sécurité : action fermée depuis
                continue
            util = joueur.action_vote(qui).utilisation_ouverte
            nom_cible = util.cible.nom if util.cible else "*non défini*"

            util.close()  # update direct pour empêcher de voter

        await chan.send(content.format(nom_cible=nom_cible))

    config.session.commit()

    # Actions déclenchées par fermeture
    if isinstance(qui, Vote):
        for action in Action.query.join(Action.base).filter(BaseAction.trigger_debut == ActionTrigger.close(qui)):
            await gestion_actions.open_action(action)

        for action in Action.query.join(Action.base).filter(BaseAction.trigger_fin == ActionTrigger.close(qui)):
            await gestion_actions.close_action(action)

    # Programme prochaine ouverture
    if heure:
        ts = tools.next_occurrence(tools.heure_to_time(heure))
        if heure_chain:
            await planif_command(ts, open_vote, qui=qui, heure=heure_chain, heure_chain=heure)
        else:
            await planif_command(ts, open_vote, qui=qui)


async def _close_action(action):
    await action.joueur.private_chan.send(
        f"{tools.montre()}  Fin de la possibilité d'utiliser ton action {tools.code(action.base.slug)} ! \n"
        f"Action définitive : {action.decision}"
    )
    await gestion_actions.close_action(action)


@close.command(name="actions")
@tools.mjs_only
@journey_command
async def close_actions(journey: DiscordJourney, heure: str):
    """Clôture les actions terminant à une heure donnée (COMMANDE BOT / MJ)

    Args:
        heure: Heure de début des actions à clôturer.

    Une sécurité empêche de fermer une action déjà en cours.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi (penser à envoyer /
    planifier la fermeture de l'action).
    """
    actions = await _get_actions("close", heure)

    await journey.send(
        "\n ".join(f" - {action.base.slug} - {action.joueur.nom}" for action in actions),
        code=True,
        prefix=f"Action(s) répondant aux critères ({len(actions)}) :",
    )
    for action in actions:
        await _close_action(action)


@close.command(name="action")
@tools.mjs_only
@journey_command
async def close_action(journey: DiscordJourney, id: int):
    """Clôture un action (COMMANDE BOT / MJ)

    Args:
        id: ID de l'action à clôturer.

    Une sécurité empêche de fermer une action déjà en cours.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi (penser à envoyer /
    planifier la fermeture de l'action).
    """
    action = await _get_action("close", id)
    if not action:
        await journey.send(f"L'action #{id} n'est pas ouverte !")
        return

    await journey.send(f"Joueur concerné : {action.joueur}")
    await _close_action(action)


remind = app_commands.Group(name="remind", description="Rappeler quelque chose")
remind = tools.mjs_only(remind)


@remind.command(name="vote")
@tools.mjs_only
@journey_command
async def remind_vote(journey: DiscordJourney, *, qui: Vote):
    """Envoi un rappel de vote / actions de rôle (COMMANDE BOT / MJ)

    Args:
        qui: Type de vote à rappeler (vote pour le condamné du jour, le nouveau maire ou la victime des loups).

    Le bot n'envoie un message qu'aux joueurs n'ayant pas encore voté / agi,
    si le vote ou l'action est bien en cours.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi !

    Example:
        - ``/remind maire`` :       rappelle le vote maire maintenant
    """
    joueurs = await _get_joueurs("remind", qui)

    await journey.send(
        "\n".join(f" - {joueur.nom}" for joueur in joueurs),
        code=True,
        prefix=f"Joueur(s) répondant aux critères ({len(joueurs)}) :",
    )

    for joueur in joueurs:
        match qui:
            case Vote.cond:
                await joueur.private_chan.send(
                    f"⏰ {joueur.member.mention} Plus que 30 minutes pour voter pour le condamné du jour ! 😱 \n"
                )
            case Vote.maire:
                await joueur.private_chan.send(
                    f"⏰ {joueur.member.mention} Plus que 30 minutes pour élire le nouveau maire ! 😱 \n"
                )
            case Vote.loups:
                await joueur.private_chan.send(
                    f"⏰ {joueur.member.mention} Plus que 30 minutes pour voter pour la victime du soir ! 😱 \n"
                )


async def _remind_action(action):
    return await action.joueur.private_chan.send(
        f"⏰ {action.joueur.member.mention} Plus que 30 minutes pour utiliser ton action "
        f"{tools.code(action.base.slug)} ! 😱 \n"
    )


@remind.command(name="actions")
@tools.mjs_only
@journey_command
async def remind_actions(journey: DiscordJourney, heure: str):
    """Rappelle d'utiliser les actions terminant à une heure donnée (COMMANDE BOT / MJ)

    Args:
        heure: Heure de début des actions à rappeler.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi !
    """
    actions = await _get_actions("remind", heure)

    await journey.send(
        "\n".join(f" - {action.base.slug} - {action.joueur.nom}" for action in actions),
        code=True,
        prefix=f"Action(s) répondant aux critères ({len(actions)}) :",
    )
    for action in actions:
        await _remind_action(action)


@remind.command(name="action")
@tools.mjs_only
@journey_command
async def remind_action(journey: DiscordJourney, id: int):
    """Rappelle d'utiliser une action précise (COMMANDE BOT / MJ)

    Args:
        id: ID de l'action à rappeler.

    Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
    Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi !
    """
    action = await _get_action("remind", id)
    if not action:
        await journey.send(f"L'action #{id} n'est pas ouverte !")
        return

    await journey.send(f"Joueur concerné : {action.joueur}")
    await _remind_action(action)


RefillMotif = enum.Enum("RefillMotif", config.refills_full + config.refills_one)


@app_commands.command()
@tools.mjs_only
@journey_command
async def refill(
    journey: DiscordJourney,
    motif: RefillMotif,
    *,
    joueur: app_commands.Transform[Joueur, tools.VivantTransformer] | None = None,
):
    """Recharger un/des pouvoirs rechargeables (COMMANDE BOT / MJ)

    Args:
        motif: Raison de rechargement (divin = forcer le refill car les MJs tout-puissants l'ont décidé).
        joueur: Si omis, recharge TOUS les joueurs.
    """
    motif: str = motif.name

    if motif in config.refills_divins:
        query = Action.query.filter(Action.active == True, Action.charges != None)
    else:
        query = Action.query.join(Action.base).filter(Action.active == True, BaseAction.refill.contains(motif))

    if joueur:
        query = query.filter(Action.joueur == joueur)
    else:
        await journey.ok_cancel("Tu as choisi de recharger le pouvoir de TOUS les joueurs actifs, en es-tu sûr ?")

    # do refill
    refillable = query.all()
    await tools.log(refillable, code=True, prefix=f"Refill {motif} {joueur.nom if joueur else 'ALL'} :")

    if joueur and len(refillable) > 1:
        action = await journey.select(
            "Action(s) répondant aux critères :",
            {action: f"{action.base.slug}, id = {action.id} \n" for action in refillable},
            placeholder="Choisir l'action à recharger",
        )
        refillable = [action]

    for action in refillable:
        await _do_refill(motif, action)
    await journey.send("Fait.")


@app_commands.command()
@tools.mjs_only
@journey_command
async def cparti(journey: DiscordJourney):
    """Lance le jeu (COMMANDE MJ)

    - Programme les votes condamnés quotidiens (avec chaînage) 10h-18h
    - Programme un vote maire 10h-18h
    - Programme les actions au lancement du jeu (choix de mentor...) et permanentes (forgeron)... à 19h
    - Crée les "actions de vote", sans quoi /open plante

    À utiliser le jour du lancement après 10h (lance les premières actions le soir et les votes le lendemain)
    """
    await journey.ok_cancel(
        "C'est parti ?\n"
        "Les rôles ont bien été attribués et synchronisés ? (si non, le faire AVANT de valider)\n\n"
        "On est bien après 10h le jour du lancement ?\n\n"
        "Tu es conscient que tous les joueurs recevront à 18h55 un message "
        "en mode « happy Hunger Games » ? (codé en dur parce que flemme)"
    )
    await journey.ok_cancel(
        "Les actions des joueurs ont été attribuées à la synchronisation des rôles, mais les /open "
        "n'ont aucun impact tant que tout le monde est en `role_actif == False` sur le Tableau de bord.\n"
        "Il faut donc **passer tout le monde à `True` maintenant** (puis `/sync silent`) avant de continuer."
    )
    await journey.ok_cancel(
        "Dernière chose à faire : activer le backup automatique du Tableau de bord tous les jours. "
        "Pour ce faire, l'ouvrir et aller dans `Extensions > Apps Script` puis dans le panel "
        "`Déclencheurs` à gauche (l'horloge) et cliquer sur `Ajouter un déclencheur` en bas à droite.\n\n"
        "Remplir les paramètres : `Backupfeuille`, `Head`, `Déclencheur horaire`, `Quotidien`, `Entre 1h et 2h` "
        "(pas plus tard car les votes du jour changent à 3h)."
    )

    rep = "C'est parti !\n"

    n10 = tools.next_occurrence(datetime.time(hour=10))
    n19 = tools.next_occurrence(datetime.time(hour=19))

    # Programmation votes condamnés chainés 10h-18h
    rep += "\nProgrammation des votes :\n"
    await planif_command(n10, open_vote, qui=Vote.cond, heure="18h", heure_chain="10h")
    rep += " - À 10h : /open cond 18h 10h\n"

    # Programmation votes loups chainés 19h-23h
    await planif_command(n19, open_vote, qui=Vote.loups, heure="23h", heure_chain="19h")
    rep += " - À 19h : /open loups 23h 19h\n"

    # Programmation premier vote maire 10h-17h
    await planif_command(n10, open_vote, qui=Vote.maire, heure="17h")
    rep += " - À 10h : /open maire 17h\n"

    # Programmation actions au lancement et actions permanentes
    rep += "\nProgrammation des actions start / perma :\n"
    start_perma = (
        Action.query.join(Action.base)
        .filter(BaseAction.trigger_debut.in_([ActionTrigger.start, ActionTrigger.perma]))
        .all()
    )
    for action in start_perma:
        rep += f" - À 19h : /open {action.id} (trigger_debut == {action.base.trigger_debut})\n"
        await planif_command(n19, open_action, id=action.id)

    # Programmation envoi d'un message aux connards
    rep += "\nEt, à 18h50 : /send all [message de hype oue oue c'est génial]\n"
    await planif_command(
        n19 - datetime.timedelta(minutes=10),
        communication.send,
        cibles="all",
        message=(
            "Ah {member.mention}... J'espère que tu es prêt(e), parce que la partie commence DANS 10 MINUTES !!!"
            "https://tenor.com/view/thehungergames-hungergames-thggifs-effie-gif-5114734"
        ),
    )
    await tools.log(rep, code=True)

    # Drop (éventuel) et (re-)création actions de vote
    Action.query.filter_by(base=None).delete()
    Action.add(*(Action(joueur=joueur, vote=vote) for joueur in Joueur.query.all() for vote in Vote))
    Action.query

    await journey.send(f"C'est tout bon ! (détails dans {config.Channel.logs.mention})")


@app_commands.command()
@tools.mjs_only
@journey_command
async def cfini(journey: DiscordJourney):
    """✨ Clôture le jeu (COMMANDE MJ)

    Supprime toutes les tâches planifiées, ce qui stoppe de fait le jeu.
    """
    await journey.ok_cancel(
        "C'est fini ?\n"
        "ATTENTION : Confirmer supprimera TOUTES LES TÂCHES EN ATTENTE, ce qui est compliqué à annuler !"
    )

    await journey.send("Suppression des tâches...")
    async with journey.channel.typing():
        taches = Tache.query.all()
        Tache.delete(*taches)  # On supprime et déprogramme le tout !

    await journey.send(
        "C'est tout bon !\n"
        "Dernière chose : penser à désactiver le backup automatique du Tableau de bord !. "
        "Pour ce faire, l'ouvrir et aller dans `Extensions > Apps Script` puis dans le panel "
        "`Déclencheurs` à gauche (l'horloge) et cliquer sur les trois points à droite du déclencheur > Supprimer."
    )



"""lg-rez / features / Commandes informatives

Commandes donnant aux joueurs des informations sur le jeu, leurs
actions, les joueurs en vie et morts...

"""

import discord
from discord import app_commands

from lgrez import config
from lgrez.blocs import tools, env
from lgrez.bdd import Joueur, Role, Camp, BaseAction, ActionTrigger, Vote, Action
from lgrez.blocs.journey import DiscordJourney, journey_command
from lgrez.features import gestion_actions


def _roles_list(roles: list[Role]) -> str:
    return "\n".join(
        str(role.camp.discord_emoji_or_none or "") + tools.code(f"{role.nom.ljust(25)} {role.description_courte}")
        for role in roles
        if not role.nom.startswith("(")
    )


DESCRIPTION = """Commandes pour en savoir plus sur soi et les autres"""


@app_commands.command()
@journey_command
async def roles(journey: DiscordJourney, *, role: app_commands.Transform[Role, tools.RoleTransformer] | None = None):
    """Affiche la liste des rôles / des informations sur un rôle.

    Args:
        role: Le rôle pour lequel avoir les informations détaillées (liste tous les rôles par défaut).

    Voir aussi la commande `/camps`.
    """
    if role:
        await journey.send(embed=role.embed)
        return

    roles = Role.query.filter_by(actif=True).order_by(Role.nom).all()
    await journey.send(
        f"Rôles trouvés :\n{_roles_list(roles)}\n"
        + tools.ital(f"({tools.code('/role <role>')} pour plus d'informations sur un rôle.)"),
    )


@app_commands.command()
@journey_command
async def camps(journey: DiscordJourney, *, camp: app_commands.Transform[Camp, tools.CampTransformer] | None = None):
    """Affiche la liste des camps / les rôles d'un camp.

    Args:
        camp: Le camp pour lequel avoir les informations détaillées et la liste des rôles.

    Voir aussi la commande `/roles`.
    """
    if camp:
        await journey.send(embed=camp.embed)
        await journey.send(
            f"Rôles dans ce camp :\n{_roles_list(camp.roles)}\n"
            + tools.ital(f"({tools.code('/roles <role>')} pour plus d'informations sur un rôle.)"),
        )
        return

    camps = Camp.query.filter_by(public=True).order_by(Camp.nom).all()
    await journey.send(
        "Camps trouvés :\n"
        + "\n".join(f"{camp.discord_emoji_or_none or ''} {camp.nom}" for camp in camps if not camp.nom.startswith("("))
        + "\n"
        + tools.ital(f"({tools.code('/camps <camp>')} pour plus d'informations sur un camp.)"),
    )


@app_commands.command()
@tools.mjs_only
@journey_command
async def rolede(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.JoueurTransformer]):
    """Donne le rôle d'un joueur (COMMANDE MJ)

    Args:
        joueur: Le joueur dont on veut connaître le rôle.
    """
    await journey.send(f"Rôle de {joueur.nom} : {joueur.role.nom_complet}")


@app_commands.command()
@tools.mjs_only
@journey_command
async def quiest(journey: DiscordJourney, *, role: app_commands.Transform[Role, tools.RoleTransformer]):
    """Liste les joueurs ayant un rôle donné (COMMANDE MJ)

    Args:
        role: Le rôle qu'on cherche.
    """
    joueurs = Joueur.query.filter_by(role=role).filter(Joueur.est_vivant).all()
    if joueurs:
        await journey.send(f"{role.nom_complet} : " + ", ".join(joueur.nom for joueur in joueurs))
    else:
        await journey.send(f"{role.nom_complet} : Personne.")


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def menu(journey: DiscordJourney):
    """Affiche des informations et boutons sur les votes / actions en cours.

    Le menu a une place beaucoup moins importante ici que sur Messenger, vu que tout est accessible par commandes.
    """
    joueur = Joueur.from_member(journey.member)

    rep = ""

    try:
        vaction = joueur.action_vote(Vote.cond)
    except RuntimeError:
        await journey.send("Minute papillon, le jeu n'est pas encore lancé !")
        return

    if vaction.is_open:
        rep += (
            f" - {config.Emoji.bucher}  Vote pour le bûcher en cours – vote actuel : {tools.code(vaction.decision)} "
            f":arrow_forward: Tape `/vote` pour voter\n"
        )

    vaction = joueur.action_vote(Vote.maire)
    if vaction.is_open:
        rep += (
            f" - {config.Emoji.maire}  Vote pour le maire en cours – vote actuel : {tools.code(vaction.decision)} "
            f":arrow_forward: Tape `/votemaire` pour voter\n"
        )

    vaction = joueur.action_vote(Vote.loups)
    if vaction.is_open:
        rep += (
            f" - {config.Emoji.lune}  Vote des loups en cours – vote actuel : {tools.code(vaction.decision)} "
            f":arrow_forward: Tape `/voteloups` pour voter\n"
        )

    if not rep:
        rep = "Aucun vote en cours.\n"

    actions = [ac for ac in joueur.actions_actives if ac.is_open]
    if actions:
        for action in actions:
            rep += (
                f" - {config.Emoji.action}  Action en cours : {tools.code(action.base.slug)} (id {action.id})"
                f" – décision : {tools.code(action.decision)} :arrow_forward: Tape `/action` pour agir\n"
            )
    else:
        rep += "Aucune action en cours.\n"

    await journey.send(f"––– MENU –––\n\n{rep}\n`/infos` pour voir ton rôle et tes actions, `@MJ` en cas de problème")


@app_commands.command()
@tools.vivants_only
@tools.private()
@journey_command
async def infos(journey: DiscordJourney):
    """Affiche tes informations de rôle / actions.

    Toutes les actions liées à ton rôle (et parfois d'autres) sont indiquées,
    même celles que tu ne peux pas utiliser pour l'instant (plus de charges, déclenchées automatiquement...)
    """
    joueur = Joueur.from_member(journey.member)
    rep = ""

    rep += f"Ton rôle actuel : {tools.bold(joueur.role.nom_complet)}\n"
    rep += tools.ital(f"({tools.code(f'/roles {joueur.role.nom}')} pour tout savoir sur ce rôle)")

    if joueur.actions_actives:
        rep += "\n\nActions :"
        rep += tools.code_bloc(
            "\n".join(
                f" - {action.base.slug.ljust(20)} "
                + (f"Cooldown : {action.cooldown}" if action.cooldown else action.base.temporalite).ljust(22)
                + (
                    f"   {action.charges} charge(s)"
                    + (" pour cette semaine" if "weekends" in action.base.refill else "")
                    if isinstance(action.charges, int)
                    else "Illimitée"
                )
                for action in joueur.actions_actives
            )
        )
        # Vraiment désolé pour cette immondice j'ai la flemme
    else:
        rep += "\n\nAucune action disponible."

    await journey.send(
        f"{rep}\n{tools.code('/menu')} pour voir les votes et "
        f"actions en cours, {tools.code('@MJ')} en cas de problème"
    )


def _actions_table(actions: list[Action]) -> str:
    return tools.code_bloc(
        "ID   Active  Action                   Début       Fin         CD   Charges   Refill         État\n"
        "-----------------------------------------------------------------------------------------------------\n"
        + "\n".join(
            str(action.id).ljust(5)
            + str(action.active).ljust(8)
            + action.base.slug.ljust(25)
            + str(
                action.base.heure_debut
                if action.base.trigger_debut == ActionTrigger.temporel
                else action.base.trigger_debut.name
            ).ljust(12)
            + str(
                action.base.heure_fin
                if action.base.trigger_fin == ActionTrigger.temporel
                else action.base.trigger_fin.name
            ).ljust(12)
            + str(action.cooldown).ljust(5)
            + str(action.charges).ljust(10)
            + str(action.base.refill).ljust(15)
            + ("Ouverte" if action.is_open else "Fermée")
            for action in actions
        )
    )


@app_commands.command()
@tools.mjs_only
@journey_command
async def actions(journey: DiscordJourney, *, joueur: app_commands.Transform[Joueur, tools.VivantTransformer]):
    """Affiche et modifie les actions d'un joueur (COMMANDE MJ)

    Args:
        joueur: Le joueur dont on veut voir ou modifier les actions.
    """
    actions = [ac for ac in joueur.actions if ac.base]

    choix = await journey.select(
        (
            f"Joueur : {joueur.nom}\n"
            f"Rôle : {joueur.role.nom_complet or joueur.role}\n"
            f"Actions : {_actions_table(actions)}\n"
            "Modifier/ajouter/stop :"
        ),
        {action: f"Modifier {action.id} {action.base.slug}" for action in actions}
        | {"new": "Ajouter une action", "stop": "Stop"},
        placeholder="Action à réaliser",
    )

    if choix == "new":  # ajouter
        base_slug, cooldown, charges = await journey.modal(
            f"Nouvelle action pour {joueur.nom}",
            discord.ui.TextInput(label="Slug de la baseaction (cf Gsheet R&A)"),
            discord.ui.TextInput(label="Cooldown (nombre entier)", max_length=2, default="0"),
            discord.ui.TextInput(label="Charges (vide = illimité)", required=False),
        )
        if not (base := BaseAction.query.filter_by(slug=base_slug).one_or_none()):
            await journey.send(f"Action `{base_slug}` invalide, vérifier dans le Gsheet Rôles et actions")
            return
        cooldown = int(cooldown)
        charges = int(charges) if charges else None

        action = gestion_actions.add_action(joueur=joueur, base=base, cooldown=cooldown, charges=charges)
        await journey.send(f"Action ajoutée (id {action.id}).")
        return

    elif choix == "stop":
        await journey.send("Au revoir.")
        return

    # Modifier
    action = choix
    gsheet_id = env.load("LGREZ_ROLES_SHEET_ID")
    url = f"https://docs.google.com/spreadsheets/d/{gsheet_id}"
    choix = await journey.buttons(
        "Pour modifier les attributs de la baseaction, modifier le Gsheet et utiliser `/fillroles`.",
        {
            "active": discord.ui.Button(label="Désactiver", style=discord.ButtonStyle.gray)
            if action.active
            else discord.ui.Button(label="Activer", style=discord.ButtonStyle.success),
            "edit": discord.ui.Button(label="Modifier cooldown / charges", style=discord.ButtonStyle.blurple),
            "gsheet": discord.ui.Button(label="Modifier sur le Gsheet", style=discord.ButtonStyle.link, url=url),
            "open_close": discord.ui.Button(
                label="Fermer l'action",
                style=discord.ButtonStyle.secondary if action.active else discord.ButtonStyle.gray,
            )
            if action.is_open
            else discord.ui.Button(
                label="Ouvrir l'action",
                style=discord.ButtonStyle.success if action.active else discord.ButtonStyle.gray,
            ),
        },
    )

    match choix:
        case "active":
            action.active = not action.active

        case "edit":
            cooldown, charges = await journey.modal(
                "Modifier l'action",
                discord.ui.TextInput(
                    label="Cooldown (nombre entier)", max_length=2, default=str(action.cooldown or "")
                ),
                discord.ui.TextInput(
                    label="Charges (vide = illimité)", required=False, default=str(action.charges or "")
                ),
            )
            action.cooldown = int(cooldown)
            action.charges = int(charges) if charges else None
            action.update()

        case "open_close":
            if action.is_open:
                await gestion_actions.close_action(action)
            else:
                await gestion_actions.open_action(action)

        case _:
            await journey.send(f"C'est possible ça ? {choix}")
            return

    await journey.send(f"Nouvelles actions : {_actions_table(actions)}")


@app_commands.command()
@journey_command
async def vivants(journey: DiscordJourney):
    """Affiche la liste des joueurs vivants.

    Aussi dite : « liste des joueurs qui seront bientôt morts »
    """
    joueurs = Joueur.query.filter(Joueur.est_vivant).order_by(Joueur.nom).all()

    mess = " Joueur                     en chambre\n"
    mess += "––––––––––––––––––––––––––––––––––––––––––––––\n"
    if config.demande_chambre:
        for joueur in joueurs:
            mess += f" {joueur.nom.ljust(25)}  {joueur.chambre}\n"
    else:
        for joueur in joueurs:
            mess += f" {joueur.nom}\n"

    await journey.send(mess, code=True, prefix=f"Les {len(joueurs)} joueurs vivants sont :")


@app_commands.command()
@journey_command
async def morts(journey: DiscordJourney):
    """Affiche la liste des joueurs morts.

    Aussi dite : « liste des joueurs qui mangent leurs morts »
    """
    joueurs = Joueur.query.filter(Joueur.est_mort).order_by(Joueur.nom).all()

    if joueurs:
        mess = ""
        for joueur in joueurs:
            mess += f" {joueur.nom}\n"
    else:
        mess = "Toi (mais tu ne le sais pas encore)"

    await journey.send(mess, code=True, prefix=f"Les {len(joueurs) or ''} morts sont :")



"""lg-rez / LGBot

Commandes"""

import traceback
import types
import typing

import discord
from discord import app_commands
import inspect
import readycheck

from lgrez import config, bdd
from lgrez.blocs import tools
from lgrez.blocs.journey import DiscordJourney
from lgrez.commons import UserInputError, CommandAbortedError
from lgrez.features import (
    actions_publiques,
    annexe,
    chans,
    communication,
    gestion_ia,
    informations,
    open_close,
    special,
    sync,
    taches,
    voter_agir,
)

if typing.TYPE_CHECKING:
    from lgrez.bot import LGBot


def _showexc(exc: Exception) -> str:
    return f"{type(exc).__name__}: {exc}"


class LGCommandTree(app_commands.CommandTree):
    def __init__(self, client: "LGBot", guild_id: int) -> None:
        super().__init__(client)
        self.guild = discord.Object(id=guild_id)
        self.enabled_commands: dict[str, app_commands.Command | app_commands.Group] = {}
        self.enabled_commands_and_subcommands: dict[str, app_commands.Command] = {}
        self.disabled_commands: dict[str, app_commands.Command | app_commands.Group] = {}

    def register_commands(self) -> None:
        # Commandes joueur : information, actions privés et publiques
        self._add_module_commands(informations)
        self._add_module_commands(voter_agir)
        self._add_module_commands(actions_publiques)
        # # Commandes MJs : gestion votes/actions, synchro GSheets, planifications, posts et embeds...
        self._add_module_commands(open_close)
        self._add_module_commands(sync)
        self._add_module_commands(taches)
        self._add_module_commands(communication)
        # # Commandes mixtes : comportement de l'IA et trucs divers
        self._add_module_commands(gestion_ia)
        self._add_module_commands(annexe)
        self._add_module_commands(chans)
        # # Commandes spéciales, méta-commandes...
        self._add_module_commands(special)

        # Commandes désactivées de base
        # self.disable_command("vote")
        # self.disable_command("votemaire")
        # self.disable_command("voteloups")
        # self.disable_command("haro")
        # self.disable_command("candid")
        # self.disable_command("setup")

    def _add_module_commands(self, module: types.ModuleType) -> None:
        for name, value in inspect.getmembers(module):
            if isinstance(value, (app_commands.Command, app_commands.Group)):
                self.add_and_enable_command(value)

            if isinstance(value, app_commands.ContextMenu):
                self.add_command(value, guild=self.guild)

    def add_and_enable_command(
        self, command: app_commands.Command | app_commands.Group, override: bool = False
    ) -> None:
        if not command.parent:  # Main command
            self.add_command(command, guild=self.guild, override=override)
            self.enabled_commands[command.qualified_name] = command
        self.enabled_commands_and_subcommands[command.qualified_name] = command

    def enable_command(self, name: str) -> bool:
        if name in self.enabled_commands:
            return False
        if name in self.disabled_commands:
            command = self.disabled_commands.pop(name)
            self.add_command(command)
            self.enabled_commands[name] = command
            if isinstance(command, app_commands.Group):
                for subcommand in command.walk_commands():
                    self.enabled_commands_and_subcommands[subcommand.qualified_name] = subcommand
            else:
                self.enabled_commands_and_subcommands[command.qualified_name] = command
            return True
        else:
            raise LookupError(f"Commande non déclarée : {name}")

    def disable_command(self, name: str) -> bool:
        if name in self.disabled_commands:
            return False
        if name in self.enabled_commands:
            command = self.enabled_commands.pop(name)
            self.remove_command(name, guild=self.guild)
            self.disabled_commands[name] = command
            if isinstance(command, app_commands.Group):
                for subcommand in command.walk_commands():
                    self.enabled_commands_and_subcommands.pop(subcommand.qualified_name, None)
            else:
                self.enabled_commands_and_subcommands.pop(command.qualified_name, None)
            return True
        else:
            raise LookupError(f"Commande non déclarée : {name}")

    def get_command_by_name(self, name: str) -> app_commands.Command | None:
        return self.enabled_commands_and_subcommands.get(name)

    async def on_error(self, interaction: discord.Interaction, error: app_commands.AppCommandError) -> None:
        """Méthode appelée par Discord à un exception dans une commande.

        Analyse l'erreur survenue et informe le joueur de manière
        adéquate en fonction, en mentionnant les MJs si besoin.

        Args:
            interaction: Interaction ayant causé l'exception
            error: Exception levée
        """
        if interaction.guild != config.guild:
            return

        message = "Strange Exception"

        running_interaction = DiscordJourney.RUNNING_INTERACTION or interaction
        DiscordJourney.RUNNING_INTERACTION = None

        match error:
            case CommandAbortedError():
                return  # Already handled (button change)

            case app_commands.CommandInvokeError():
                if isinstance(error.original, (bdd.SQLAlchemyError, bdd.DriverOperationalError)):  # Erreur BDD
                    try:
                        config.session.rollback()  # On rollback la session
                        await tools.log("Rollback session")
                    except readycheck.NotReadyError:
                        pass

                # Dans tous les cas, si erreur à l'exécution
                command = interaction.command.name
                prefix = f":x: Oups ! Un problème est survenu à l'exécution de la commande `{command}` :grimacing: :"

                if not config.is_setup or (
                    isinstance(interaction.user, discord.Member) and interaction.user.top_role >= config.Role.mj
                ):
                    # MJ / webhook : affiche le traceback complet
                    tb = traceback.format_exception(type(error.original), error.original, error.original.__traceback__)
                    await tools.send_blocs(running_interaction, "".join(tb), code=True, prefix=prefix)
                    return
                else:
                    # Pas MJ : exception seulement
                    message = f"{prefix}\n{config.Role.mj.mention} ALED – " + tools.ital(_showexc(error.original))

            case app_commands.CommandNotFound():
                message = (
                    ":x: Hum, je ne connais pas cette commande  :thinking:\n"
                    "Tape `/` puis clique sur une des commandes de la liste."
                )

            case app_commands.MissingRole() if error.missing_role == config.Role.mj.name:
                message = ":x: Hé ho toi, cette commande est réservée aux MJs !  :angry:"

            case app_commands.MissingRole() if error.missing_role == config.Role.joueur_en_vie.name:
                message = ":x: Désolé, cette commande est réservée aux joueurs en vie !"

            case app_commands.MissingAnyRole():
                message = (
                    ":x: Cette commande est réservée aux joueurs ! (parce qu'ils doivent être inscrits en base) "
                    f"({tools.code('/doas')} est là en cas de besoin)"
                )

            case UserInputError():
                # Autre check non vérifié
                message = f':x: Argument "{error.param}" invalide : {error.message}'

            case app_commands.CheckFailure():
                # Autre check non vérifié
                if interaction.response.is_done():  # Déjà géré
                    return
                message = (
                    ":x: Tiens, il semblerait que cette commande ne puisse pas être exécutée ! "
                    f"{tools.mention_MJ(interaction)} ?\n({tools.ital(_showexc(error))})"
                )

            case _:
                message = (
                    f":x: Oups ! Une erreur inattendue est survenue  :grimacing:\n"
                    f"{tools.mention_MJ(interaction)} ALED – {tools.ital(_showexc(error))}"
                )

        await tools.send_blocs(interaction, message)



"""lg-rez / features / Process d'inscription

Étapes préliminaires pour les joueurs avant le début de saison

"""

import asyncio

import discord

from lgrez import config
from lgrez.blocs import tools, env, gsheets
from lgrez.bdd import Joueur, Role, Camp, Statut
from lgrez.blocs.journey import DiscordJourney


async def new_channel(member: discord.Member) -> discord.TextChannel:
    """Crée un nouveau salon privé.

    Peut être étendue, mais toujours appeler cette fonction pour créer le chan en lui-même,
    au risque d'altérer le fonctionnement normal du bot.

    Args:
        member: Le membre pour qui créer le salon.

    Returns:
        Le salon créé.
    """
    categ = await tools.multicateg(config.private_chan_category_name)
    chan = await member.guild.create_text_channel(
        f"{config.private_chan_prefix}{member.name}",
        topic=str(member.id),  # topic provisoire : ID du membre
        category=categ,
    )
    await chan.set_permissions(member, read_messages=True, send_messages=True)

    return chan


async def register_on_tdb(joueur: Joueur) -> None:
    """Enregistre un joueur dans le Tableau de bord.

    Peut être personnalisé à un autre système de gestion des joueurs.

    Args:
        joueur: Le joueur à enregistrer.

    Note:
        Fonction asynchrone depuis la version 2.2.2.
    """
    SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")
    workbook = await gsheets.connect(SHEET_ID)
    sheet = await workbook.worksheet(config.tdb_main_sheet)
    values = await sheet.get_all_values()  # Liste de listes

    head = values[config.tdb_header_row - 1]
    # Ligne d'en-têtes (noms des colonnes), - 1 car indexé à 0

    id_index = gsheets.a_to_index(config.tdb_id_column)
    pk = head[id_index]
    if pk != Joueur.primary_col.key:
        raise ValueError(
            "Tableau de bord : la cellule "
            "`config.tdb_id_column` / `config.tdb_header_row` = "
            f"`{config.tdb_id_column}{config.tdb_header_row}` "
            f"vaut `{pk}` au lieu de la clé primaire de la table "
            f"`Joueur`, `{Joueur.primary_col.key}` !"
        )

    mstart, mstop = config.tdb_main_columns
    main_indexes = range(gsheets.a_to_index(mstart), gsheets.a_to_index(mstop) + 1)
    # Indices des colonnes à remplir
    cols = {}
    for index in main_indexes:
        col = head[index]
        if col in Joueur.attrs:
            cols[col] = Joueur.attrs[col]
        else:
            raise ValueError(
                f"Tableau de bord : l'index de la zone principale "
                f"`{col}` n'est pas une colonne de la table `Joueur` !"
                " (voir `lgrez.config.main_indexes` / "
                "`lgrez.config.tdb_header_row`)"
            )

    tstart, tstop = config.tdb_tampon_columns
    tampon_indexes = range(gsheets.a_to_index(tstart), gsheets.a_to_index(tstop) + 1)

    TDB_tampon_index = {}
    for index in tampon_indexes:
        col = head[index].partition("_")[2]
        if col in cols:
            TDB_tampon_index[col] = index
        else:
            raise ValueError(
                f"Tableau de bord : l'index de zone tampon `{head[index]}` "
                f"réfère à la colonne `{col}` (partie suivant le premier "
                f"underscore), qui n'est pas une colonne de la zone "
                "principale ! (voir `lgrez.config.tampon_indexes` / "
                "`lgrez.config.main_indexes`)"
            )

    plv = config.tdb_header_row  # Première Ligne Vide
    # (si tableau vide, suit directement le header, - 1 pour l'indexage)
    for i_row, row in enumerate(values):
        # On parcourt les lignes du TDB
        if i_row < config.tdb_header_row:
            # Ligne avant le header / le header (car décalage de 1)
            continue

        if row[id_index].isdigit():
            # Si il y a un vrai ID dans la colonne ID
            plv = i_row + 1

    modifs = [gsheets.Modif(plv, id_index, joueur.discord_id)]
    for index in main_indexes:
        val = getattr(joueur, head[index])
        modifs.append(gsheets.Modif(plv, index, val))
    for index in tampon_indexes:
        val = getattr(joueur, head[index].partition("_")[2])
        # Colonnes "tampon_<col>" ==> <col>
        modifs.append(gsheets.Modif(plv, index, val))

    try:
        await gsheets.update(sheet, *modifs)
    except gsheets.ConnectionError:
        await joueur.private_chan.send("Petit problème, on réessaie...")
        await asyncio.sleep(10)
        await gsheets.update(sheet, *modifs)


# Routine d'inscription (fonction appelée par la commande /co)
async def main(member: discord.Member) -> None:
    """Routine d'inscription complète d'un joueur.

    Args:
        member: Le joueur à inscrire.

    Crée et paramètre le salon privé du joueur, lui pose des questions et l'inscrit en base.

    Personnalisation : voir :obj:`.config.demande_chambre`, :obj:`.config.chambre_mj`,
    :func:`.config.additional_inscription_step` et :obj:`.config.debut_saison`.

    Commande appelée à l'arrivée sur le serveur, utiliser :meth:`\/co <.bot.Special.Special.co.callback>`
    pour trigger cette commande depuis Discord.
    """
    try:
        joueur = Joueur.from_member(member)
    except ValueError:  # Joueur pas encore inscrit en base
        pass
    else:  # Joueur dans la bdd = déjà inscrit
        chan = joueur.private_chan
        await chan.set_permissions(member, read_messages=True, send_messages=True)
        await chan.send(f"Saloww ! {member.mention} tu es déjà inscrit, viens un peu ici !")
        return

    if chan := next((chan for chan in config.guild.text_channels if chan.topic == str(member.id)), None):
        # Inscription en cours (topic du chan = ID du membre)
        await chan.set_permissions(member, read_messages=True, send_messages=True)
        await chan.send(f"Tu as déjà un salon à ton nom, {member.mention}, par ici !")

    else:  # Pas d'inscription déjà en cours : création channel
        chan = await new_channel(member)

    # Récupération nom et renommages

    await chan.send(
        f"Bienvenue {member.mention} ! Je suis le bot à qui tu auras affaire tout au long de la partie.\n"
        "Tu es ici sur ton salon textuel privé, auquel seul toi et les MJ ont accès ! "
        "C'est ici que tu lanceras toutes les commandes pour voter, te renseigner... "
        "mais aussi que les MJ discuteront avec toi en cas de soucis.\n\n"
        "Pas de panique, je vais tout t'expliquer !"
    )

    class InscriptionView(discord.ui.View):
        @discord.ui.button(label="Terminer mon inscription", style=discord.ButtonStyle.success)
        async def callback(self, interaction: discord.Interaction, button: discord.ui.Button):
            async with DiscordJourney(interaction) as journey:
                await _inscription_process(journey, member, chan)

        async def on_error(self, _interaction: discord.Interaction, error: Exception, _item: discord.ui.Item) -> None:
            raise error

    await chan.send("Avant toute chose, finalisons ton inscription !", view=InscriptionView(timeout=None))


async def _inscription_process(journey: DiscordJourney, member: discord.Member, chan: discord.TextChannel) -> None:
    await journey.send(
        "Très bien !\nD'abord, un point règles nécessaire :\n\n"
        + tools.quote_bloc(
            "En t'inscrivant au Loup-Garou de la Rez, tu garantis vouloir participer à cette édition et "
            "t'engages à respecter les règles du jeu. Si tu venais à entraver le bon déroulement de la partie "
            "pour une raison ou une autre, les MJ s'autorisent à te mute ou t'expulser du Discord sans préavis."
        )
    )

    await tools.sleep(chan, 5)

    if not await journey.yes_no("C'est bon pour toi ?"):
        await journey.send(
            f"Pas de soucis. Si tu changes d'avis ou que c'est un missclick, appelle un MJ aled ({tools.code('@MJ')})."
        )
        return

    await journey.buttons(
        "Parfait. Je vais d'abord avoir besoin de ton (vrai) nom, celui par lequel on t'appelle au quotidien. "
        "Attention, tout troll sera foudracanonné™ !",
        {None: discord.ui.Button(label="Go !", style=discord.ButtonStyle.success)},
    )

    ok = False
    while not ok:
        prenom, nom_famille = await journey.modal("À qui avons-nous l'honneur ?", "Prénom", "Nom")
        nom = f"{prenom.strip().title()} {nom_famille.strip().title()}"

        ok = await journey.yes_no(
            f"Tu me dis donc t'appeler {tools.bold(nom)}. C'est bon pour toi ? Pas d'erreur, pas de troll ?"
        )

    await chan.edit(name=config.private_chan_prefix + nom)  # Renommage conv
    if member.top_role < config.Role.mj:
        # Renommage joueur (ne peut pas renommer les MJ)
        await member.edit(nick=nom)

    await journey.send(
        "Parfait ! Je t'ai renommé(e) pour que tout le monde te reconnaisse, et j'ai renommé cette conversation."
    )

    if config.demande_chambre:
        await tools.sleep(chan, 3)
        if await journey.yes_no("Bien, dernière chose : habites-tu à la Rez ?"):
            (chambre,) = await journey.modal("Et où te trouve-t-on ?", "Numéro de chambre")
        else:
            chambre = config.chambre_mj
    else:
        chambre = None

    res = await config.additional_inscription_step(journey, member, chan)
    if res is False:
        return

    if config.demande_chambre:
        await journey.send(f"{nom}, en chambre {chambre}... Je t'inscris en base !")
    else:
        await journey.send(f"{nom}... Je t'inscris en base !")

    async with chan.typing():
        # Enregistrement en base

        joueur = Joueur(
            discord_id=member.id,
            chan_id_=chan.id,
            nom=member.display_name,
            chambre=chambre,
            statut=Statut.vivant,
            role=Role.default(),
            camp=Camp.default(),
            votant_village=True,
            votant_loups=False,
            role_actif=False,
        )
        joueur.add()

        # Ajout sur le TDB

        await register_on_tdb(joueur)

        # Grant accès aux channels joueurs et information

        await member.add_roles(config.Role.joueur_en_vie)
        await chan.edit(topic="Ta conversation privée avec le bot, c'est ici que tout se passera !")

    # Conseiller d'ACTIVER TOUTES LES NOTIFS du chan
    # (et mentions only pour le reste, en activant @everyone)
    await chan.send(
        f"Tu es maintenant inscrit(e) ! Je t'ai attribué le rôle {config.Role.joueur_en_vie.mention}, "
        "qui te donne l'accès à tout plein de nouveaux salons à découvrir."
    )
    await chan.send(
        "Juste quelques dernières choses :\n "
        "- Plein de commandes te sont d'ores et déjà accessibles ! "
        f"Découvre les toutes en tapant `/` dans la zone d'envoi de messages ;\n"
        f"- Si tu as besoin d'aide, mentionne simplement les MJs ({tools.code('@' + config.Role.mj.name)}) "
        "et on viendra voir ce qui se passe !\n "
        "- Si ce n'est pas le cas, je te conseille fortement d'installer Discord sur ton téléphone, et "
        f"d'{tools.bold('activer toutes les notifications')} pour ce salon ! Promis, pas de spam :innocent:\n"
        'Pour le reste du serveur, tu peux le mettre en "mentions only",'
        f"en activant le {tools.code('@everyone')} – il est limité ;\n\n"
        "Enfin, n'hésite pas à me parler, j'ai toujours quelques réponses en stock..."
    )

    await tools.sleep(chan, 5)
    await chan.send(
        f"Voilà, c'est tout bon ! Installe toi bien confortablement, la partie commence le {config.debut_saison}."
    )

    # Log
    await tools.log(
        f"Inscription de {member.name}#{member.discriminator} réussie\n"
        f" - Nom : {nom}\n - Chambre : {chambre}\n"
        f" - Channel créé : {chan.mention}"
    )



import re
import setuptools


repo_url = "https://github.com/GRI-ESPCI/lg-rez"

version = ""
with open("lgrez/__init__.py") as f:
    version = re.search(r"""^__version__\s*=\s*['"](.*?)['"]""", f.read(), re.MULTILINE).group(1)

assert version, "__version__ not found in __init__.py"


def absolute_links(string, base):
    """Replace all relative Markdown links by absolute links"""
    base = base.rstrip("/")
    return re.sub(
        r"\[(.+?)\]\(([^:]+?)\)",  # Every [text](link) without ":" in link
        f"[\\1]({base}/\\2)",  # Replace link by base/link
        string,
    )


with open("README.md", "r", encoding="utf-8") as fh:
    readme = fh.read()

version_blob_url = f"{repo_url}/blob/{version}"
# Github blob URL for tag <version>
long_description = absolute_links(readme, version_blob_url)
# long_description is README.md contents with relative links to project
# files/folders converted into links to files on GitHub blob, so they
# can be clicked on PyPI.


with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.readlines()


setuptools.setup(
    name="lg-rez",
    version=version,
    author="Loïc Simon, Tom Lacoma",
    author_email="loic.simon@espci.org, tom.lacoma@espci.org",
    description="Discord bot for organizing Werewolf RP games ESPCI-style",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=repo_url,
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Natural Language :: French",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Games/Entertainment :: Role-Playing",
        "Topic :: Internet",
    ],
    install_requires=requirements,
    python_requires=">=3.10",
    package_data={
        "lgrez": ["server_structure.json"],
    },
    include_package_data=True,
)



"""Loup-Garou de la Rez (lg-rez)

Discord bot for organizing boisterous Werewolf RP games ESPCI-style.

See github.com/GRI-ESPCI/lg-rez for informations.
"""

__title__ = "lg-rez"
__author__ = "Loïc Simon, Tom Lacoma"
__license__ = "MIT"
__copyright__ = "Copyright 2020 - 2022 Loïc Simon & Tom Lacoma - Club BD-Jeux × GRIs – ESPCI Paris - PSL"
__version__ = "4.0.7"
__all__ = ["LGBot"]


from lgrez import bot

LGBot = bot.LGBot  # Accès direct à lgrez.LGBot



"""lg-rez / Gestion des données

Déclaration de toutes les tables, colonnes, méthodes, et connection à la BDD

"""

from sqlalchemy.exc import SQLAlchemyError, OperationalError as DriverOperationalError

from lgrez.bdd import base
from lgrez.bdd.base import connect, tables
from lgrez.bdd.enums import Statut, ActionTrigger, CandidHaroType, Vote, UtilEtat, CibleType
from lgrez.bdd.model_joueurs import Joueur, CandidHaro, Boudoir, Bouderie
from lgrez.bdd.model_jeu import Role, Camp, BaseAction, BaseCiblage
from lgrez.bdd.model_actions import Action, Utilisation, Ciblage, Tache
from lgrez.bdd.model_ia import Reaction, Trigger

# All tables and enums directly accessible via bdd.<name>


#: __all__ = toutes les classes de données publiques
__all__ = [nom for nom in base.tables if not nom.startswith("_")]



"""lg-rez / features / Communication

Envoi de messages, d'embeds...

"""
import datetime
import functools
import os
import re
from typing import Literal

import discord
from discord import app_commands
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from lgrez import commons, config
from lgrez.blocs import tools, gsheets
from lgrez.blocs.journey import DiscordJourney, journey_command, journey_context_menu
from lgrez.bdd import (
    Joueur,
    Action,
    Camp,
    BaseAction,
    Utilisation,
    Statut,
    ActionTrigger,
    CandidHaroType,
    UtilEtat,
    Vote,
)
from lgrez.features import gestion_actions
from lgrez.features.sync import transtype


def _joueur_repl(mtch: re.Match) -> str:
    """Remplace @... par la mention d'un joueur, si possible"""
    nearest = Joueur.find_nearest(mtch.group(1), col=Joueur.nom, sensi=0.8)
    if nearest:
        joueur = nearest[0][0]
        try:
            return joueur.member.mention
        except ValueError:
            pass
    return mtch.group(0)


def _role_repl(mtch: re.Match) -> str:
    """Remplace @role_slug par la mention du rôle, si possible"""
    try:
        role = getattr(config.Role, mtch.group(1).lower())
    except AttributeError:
        return mtch.group(0)
    else:
        return role.mention


def _emoji_repl(mtch: re.Match) -> str:
    """Remplace :(emoji): par la représentation de l'emoji, si possible"""
    emo = tools.emoji(mtch.group(1), must_be_found=False)
    if emo:
        return str(emo)
    return mtch.group()


async def _chose_role_and_camp(journey: DiscordJourney, joueur: Joueur) -> tuple[str, discord.Emoji | None]:
    role = joueur.role.nom_complet
    if await journey.yes_no(f"Rôle à afficher pour {joueur.nom} = {role} ? (Pas double peau ou autre)"):
        emoji_camp = joueur.camp.discord_emoji_or_none
    else:
        (role,) = await journey.modal(
            "Choix du rôle à afficher",
            discord.ui.TextInput(label=f"Rôle de {joueur.nom} = {role}"[:45]),
        )
        camp = await journey.select(
            f'Camp du rôle "{role}" :',
            {
                camp: discord.SelectOption(label=camp.nom, emoji=camp.discord_emoji)
                for camp in Camp.query.filter_by(public=True).all()
            },
        )
        emoji_camp = camp.discord_emoji
    return role, emoji_camp


DESCRIPTION = """Commandes d'envoi de messages, d'embeds, d'annonces..."""

current_embed = None


@app_commands.command()
@tools.mjs_only
@journey_command
async def send(journey: DiscordJourney, *, cibles: str, message: str):
    """Envoie un message à tous ou certains joueurs (COMMANDE MJ)

    Args:
        cibles: Destinataires (`all`, `vivants`, `morts`, `<crit>=<filtre>`, `<joueur`)
        message: Message, éventuellement formaté ({member.mention}, ...)

    ``cibles`` peut être :
        - ``all`` :             Tous les joueurs inscrits, vivants et morts
        - ``vivants`` :         Les joueurs en vie
        - ``morts`` :           Les joueurs morts
        - ``<crit>=<filtre>`` : Les joueurs répondant au critère ``Joueur.<crit> == <filtre>``.
            ``crit`` peut être ``"nom"``, ``"chambre"``, ``"statut"``, ``"role"``, ``"camp"``...
            L'ensemble doit être entouré de guillemets si ``filtre`` contient un espace.
            Les rôles/camps sont cherchés par slug.
        - *le nom d'un joueur*  (raccourci pour ``nom=X``, doit être entouré de guillemets si nom + prénom)

    ``message`` peut contenir un ou plusieurs bouts de code Python à évaluer, entourés d'accolades.

    L'évaluation est faite séparément pour chaque joueur, ce qui permet de personnaliser le message
    grâce aux variables particulières dépendant du joueur :

        - ``joueur`` :  objet BDD du joueur recevant le message ==> ``joueur.nom``, ``joueur.role``...
        - ``member`` :  objet :class:`discord.Member` associé ==> ``member.mention``
        - ``chan``   :  objet :class:`discord.TextChannel` du chan privé du joueur

    Les différentes tables de données sont accessibles sous leur nom (``Joueur``, ``Role``...)

    Il est impossible d'appeler des coroutines (await) dans le code à évaluer.

    Examples:
        - ``/send all Bonsoir à tous c'est Fanta``
        - ``/send vivants Attention {member.mention}, derrière toi c'est affreux !``
        - ``/send "role=servante" Ça va vous ? Vous êtes bien {joueur.role.nom} ?``
    """
    if cibles == "all":
        joueurs = Joueur.query.all()
    elif cibles == "vivants":
        joueurs = Joueur.query.filter(Joueur.est_vivant).all()
    elif cibles == "morts":
        joueurs = Joueur.query.filter(Joueur.est_mort).all()
    elif "=" in cibles:
        crit, _, filtre = cibles.partition("=")
        crit = crit.strip()
        if crit in Joueur.attrs:
            col = Joueur.attrs[crit]
            arg = transtype(filtre.strip(), col)
            joueurs = Joueur.query.filter_by(**{crit: arg}).all()
        else:
            raise commons.UserInputError(f"critère '{crit}' incorrect")
    else:
        raise commons.UserInputError("cibles", f"Destinataires invalides : {cibles}")

    descr = f"Cibles : {cibles}\nMessage : {message}\n"

    if not joueurs:
        await journey.send(f"{descr}:x: Aucun joueur trouvé.")
        return

    await journey.send(f"{descr}:arrow_forward: {len(joueurs)} trouvé(s), envoi...")
    for joueur in joueurs:
        member = joueur.member
        chan = joueur.private_chan

        evaluated_message = tools.eval_accols(message, locals_=locals())
        await chan.send(evaluated_message)

    await journey.channel.send("Fini.")


@app_commands.command()
@tools.mjs_only
@journey_command
async def post(journey: DiscordJourney, *, chan: discord.TextChannel, message: str):
    """Envoie un message dans un salon (COMMANDE MJ)

    Args:
        chan: Salon ou poster le message.
        message: Message à envoyer (utiliser "\n" pour un saut de ligne).
    """
    message = message.replace(r"\n", "\n")
    await chan.send(message)
    await journey.send(f"Message posté sur {chan.mention} :\n" + tools.quote_bloc(message))


@app_commands.command()
@tools.mjs_only
@journey_command
async def plot(journey: DiscordJourney, *, quoi: Literal["cond", "maire"], depuis: str | None = None):
    """Trace le résultat du vote et l'envoie sur #annonces (COMMANDE MJ)

    Args:
        quoi: Vote pour le condamné ou pour l'élection à la Mairie ?
        depuis: Heure à partir de laquelle compter les votes (si plusieurs votes dans la journée, HHh / HHhMM).
            Compte tous les votes du jour par défaut.
            Si plus tard que l'heure actuelle, compte les votes de la veille.

    Trace les votes sous forme d'histogramme à partir du Tableau de bord, en fait un embed
    en précisant les résultats détaillés et l'envoie sur le chan ``#annonces``.

    Si ``quoi == "cond"``, déclenche aussi les actions liées au mot des MJs (:attr:`.bdd.ActionTrigger.mot_mjs`).
    """
    # Différences plot cond / maire
    if quoi == "cond":
        vote_enum = Vote.cond
        haro_candidature = CandidHaroType.haro
        typo = "bûcher du jour"
        mort_election = "Mort"
        pour_contre = "contre"
        emoji = config.Emoji.bucher
        couleur = 0x730000

    else:
        vote_enum = Vote.maire
        haro_candidature = CandidHaroType.candidature
        typo = "nouveau maire"
        mort_election = "Élection"
        pour_contre = "pour"
        emoji = config.Emoji.maire
        couleur = 0xD4AF37

    if depuis:
        tps = tools.heure_to_time(depuis)
    else:
        tps = datetime.time(0, 0)

    ts = datetime.datetime.combine(datetime.date.today(), tps)
    if ts > datetime.datetime.now():  # hier
        ts -= datetime.timedelta(days=1)

    log = f"/plot {quoi} (> {ts}) :"
    query = Utilisation.query.join(Utilisation.action).filter(
        Utilisation.etat == UtilEtat.validee,
        Utilisation.ts_decision > ts,
        Action.active == True,
    )
    cibles = {}

    # Get votes
    utils = query.join(Utilisation.action).filter(Action.vote == vote_enum).all()
    votes = {util.action.joueur: util.cible for util in utils}
    votelog = " / ".join(f"{v.nom} -> {c.nom}" for v, c in votes.items())
    log += f"\n  - Votes : {votelog}"

    for votant, vote in votes.items():
        cibles.setdefault(vote, [])
        cibles[vote].append(votant.nom)

    # Get intriguants
    intba = BaseAction.query.get(config.modif_vote_baseaction)
    if intba:
        log += "\n  - Intrigant(s) : "
        for util in query.join(Utilisation.action).filter(Action.base == intba).all():

            votant = util.ciblage("cible").valeur
            vote = util.ciblage("vote").valeur
            log += f"{util.action.joueur.nom} : {votant.nom} -> {vote.nom} / "

            initial_vote = votes.get(votant)
            if initial_vote:
                cibles[initial_vote].remove(votant.nom)
                if not cibles[initial_vote]:  # plus de votes
                    del cibles[initial_vote]
            votes[votant] = vote
            cibles.setdefault(vote, [])
            cibles[vote].append(votant.nom)

    # Tri des votants
    for votants in cibles.values():
        votants.sort()  # ordre alphabétique

    # Get corbeaux, après tri -> à la fin
    corba = BaseAction.query.get(config.ajout_vote_baseaction)
    if corba:
        log += "\n  - Corbeau(x) : "
        for util in query.join(Utilisation.action).filter(Action.base == corba).all():
            log += f"{util.action.joueur.nom} -> {util.cible} / "
            cibles.setdefault(util.cible, [])
#            cibles[util.cible].extend([util.action.joueur.role.nom] * config.n_ajouts_votes) #ancienne version 
            cibles[util.cible].extend(["Corbeau"] * config.n_ajouts_votes) #fix car pb avec imprimeur
    
    impri = BaseAction.query.get("dépôt-affiche")
    if impri:
        log += "\n  - Imprimante(x) : "
        for util in query.join(Utilisation.action).filter(Action.base == impri).all():
            log += f"{util.action.joueur.nom} -> {util.cible} / "
            cibles.setdefault(util.cible, [])
            cibles[util.cible].extend(["Imprimeur"]* 1)

    # Classe utilitaire
    @functools.total_ordering
    class _Cible:
        """Représente un joueur ciblé, pour usage dans /plot"""

        def __init__(self, joueur, votants):
            self.joueur = joueur
            self.votants = votants

        def __repr__(self) -> str:
            return f"{self.joueur.nom} ({self.votes})"

        def __eq__(self, other):
            if not isinstance(other, type(self)):
                return NotImplemented
            return self.joueur.nom == other.joueur.nom and self.votes == other.votes

        def __lt__(self, other):
            if not isinstance(other, type(self)):
                return NotImplemented
            if self.votes == other.votes:
                return self.joueur.nom < other.joueur.nom
            return self.votes < other.votes

        @property
        def votes(self):
            return len(self.votants)

        @property
        def eligible(self):
            return any(ch.type == haro_candidature for ch in self.joueur.candidharos)

        def couleur(self, choisi) -> str:
            if self == choisi:
                return hex(couleur).replace("0x", "#")
            if self.eligible:
                return "#64b9e9"
            else:
                return "gray"

    # Récupération votes
    cibles = [_Cible(jr, vts) for (jr, vts) in cibles.items()]
    cibles.sort(reverse=True)  # par nb de votes, puis ordre alpha
    log += f"\n  - Cibles : {cibles}"

    # Détermination cible
    choisi = None
    eligibles = [cible for cible in cibles if cible.eligible]
    log += f"\n  - Éligibles : {eligibles}"

    if eligibles:
        maxvotes = eligibles[0].votes
        egalites = [cible for cible in eligibles if cible.votes == maxvotes]

        if len(egalites) > 1:  # Égalité
            choisi = await journey.select(
                "Égalité entre plusieurs joueurs :" + "\nQui meurt / est élu ? "
                "(regarder vote du maire, si joueur garde-loupé ou inéligible...)",
                {cible: cible.joueur.nom for cible in egalites} | {None: "Personne (pas de vote du maire)"},
            )

        elif await journey.yes_no(
            f"Joueur éligible le plus voté : {tools.bold(eligibles[0].joueur.nom)}\n"
            "Ça meurt / est élu ? (pas garde-loupé, inéligible ou autre)"
        ):
            choisi = eligibles[0]

    log += f"\n  - Choisi : {choisi or '[aucun]'}"
    await tools.log(log)

    # Paramètres plot
    discord_gray = "#2F3136"
    plt.figure(facecolor=discord_gray)
    plt.rcParams.update({"font.size": 16})
    ax = plt.axes(facecolor="#8F9194")  # coloration de TOUT le graphe
    ax.tick_params(axis="both", colors="white")
    ax.spines["bottom"].set_color("white")
    ax.spines["left"].set_color(discord_gray)
    ax.spines["right"].set_color(discord_gray)
    ax.spines["top"].set_color(discord_gray)
    ax.set_facecolor(discord_gray)
    ax.set_axisbelow(True)
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    # Plot
    ax.bar(
        x=range(len(cibles)),
        height=[cible.votes for cible in cibles],
        tick_label=[cible.joueur.nom.replace(" ", "\n", 1) for cible in cibles],
        color=[cible.couleur(choisi) for cible in cibles],
    )
    plt.grid(axis="y")

    if not os.path.isdir("figures"):
        os.mkdir("figures")

    now = datetime.datetime.now().strftime("%Y-%m-%d--%H")
    image_path = f"figures/hist_{now}_{quoi}.png"
    plt.savefig(image_path, bbox_inches="tight")

    # --------------- Partie Discord ---------------

    # Détermination rôle et camp
    emoji_camp = None
    if choisi:
        if quoi == "cond":
            role, emoji_camp = await _chose_role_and_camp(journey, choisi.joueur)
            nom_et_role = f"{tools.bold(choisi.joueur.nom)}, {role}"
        else:
            # Maire : ne pas annoncer le rôle
            nom_et_role = f"{tools.bold(choisi.joueur.nom)}"
    else:
        nom_et_role = "personne, bande de tocards"

    # Création embed
    embed = discord.Embed(
        title=f"{mort_election} de {nom_et_role}", description=f"{len(votes)} votes au total", color=couleur
    )
    embed.set_author(name=f"Résultats du vote pour le {typo}", icon_url=emoji.url)

    if emoji_camp:
        embed.set_thumbnail(url=emoji_camp.url)

    embed.set_footer(
        text="\n".join(
            ("A" if cible.votes == 1 else "Ont")
            + f" voté {pour_contre} {cible.joueur.nom} : "
            + ", ".join(cible.votants)
            for cible in cibles
        )
    )

    file = discord.File(image_path, filename="image.png")
    embed.set_image(url="attachment://image.png")

    await journey.ok_cancel("Ça part ?", file=file, embed=embed)

    # Envoi du graphe
    file = discord.File(image_path, filename="image.png")
    # Un objet File ne peut servir qu'une fois, il faut le recréer

    await config.Channel.annonces.send("@everyone Résultat du vote ! :fire:", file=file, embed=embed)
    await journey.send(f"Et c'est parti dans {config.Channel.annonces.mention} !")

    if quoi == "cond":
        # Actions au mot des MJs
        for action in Action.query.filter(Action.base.has(trigger_debut=ActionTrigger.mot_mjs)).all():
            await gestion_actions.open_action(action)

        await journey.channel.send("(actions liées au mot MJ ouvertes)")


@app_commands.command()
@tools.mjs_only
@journey_command
async def annoncemort(
    journey: DiscordJourney,
    *,
    victime: app_commands.Transform[Joueur, tools.VivantTransformer],
    victime_2: app_commands.Transform[Joueur, tools.VivantTransformer] | None = None,
    victime_3: app_commands.Transform[Joueur, tools.VivantTransformer] | None = None,
    victime_4: app_commands.Transform[Joueur, tools.VivantTransformer] | None = None,
    victime_5: app_commands.Transform[Joueur, tools.VivantTransformer] | None = None,
):
    """Annonce un ou plusieurs mort(s) hors-vote (COMMANDE MJ)

    Args:
        victime: (Premier) mort à annoncer.
        victime_2: Deuxième mort à annoncer.
        victime_3: Troisième mort à annoncer.
        victime_4: Quatrième mort à annoncer.
        victime_5: Cinquième mort à annoncer (pour plus, faire en 2 fois, merde)

    Envoie un embed par mort dans ``#annonces``
    """
    joueurs = [joueur for joueur in (victime, victime_2, victime_3, victime_4, victime_5) if joueur]
    roles = []
    emojis = []
    for joueur in joueurs:
        role, emoji_camp = await _chose_role_and_camp(journey, joueur)

        if joueur.statut == Statut.MV:
            if await journey.yes_no("Annoncer la mort-vivance ?"):
                role += " Mort-Vivant"
            else:
                emoji_camp = joueur.role.camp.discord_emoji_or_none

        roles.append(role)
        emojis.append(emoji_camp)

    contextes = await journey.modal("Contextes", *[f"Pour {joueur.nom} :" for joueur in joueurs])
    authors = [
        "Oh mon dieu, quelqu'un est mort !",
        "Oh non, un deuxième !!",
        "Mais c'est une vraie hécatombe !?!?",
        "Cela s'arrêtera-t-il un jour ?",
        "Bon, ça va commencer à se voir...",
    ]

    # Création embeds
    embeds = []
    for joueur, role, emoji_camp, contexte, author in zip(joueurs, roles, emojis, contextes, authors):
        embed = discord.Embed(title=f"Mort de {tools.bold(joueur.nom)}, {role}", description=contexte, color=0x730000)
        embed.set_author(name=author)
        if emoji_camp:
            embed.set_thumbnail(url=emoji_camp.url)

        embeds.append(embed)

    await journey.ok_cancel("Ça part ?", embeds=embeds)
    await config.Channel.annonces.send("@everyone Il s'est passé quelque chose ! :scream:", embeds=embeds)
    await journey.send(f"Et c'est parti dans {config.Channel.annonces.mention} !")


@app_commands.command()
@tools.mjs_only
@journey_command
async def lore(journey: DiscordJourney, doc_id: str):
    """Récupère et poste un lore depuis un Google Docs (COMMANDE MJ)

    Convertit les formats et éléments suivants vers Discord :
        - Gras, italique, souligné, barré;
        - Petites majuscules (-> majuscules);
        - Polices à chasse fixe (Consolas / Courier New) (-> code);
        - Liens hypertextes;
        - Listes à puces;
        - Mentions de joueurs, sous la forme ``@Prénom Nom``;
        - Mentions de rôles, sous la forme ``@nom_du_role``;
        - Emojis, sous la forme ``:nom:``.

    Permet soit de poster directement dans #annonces, soit de
    récupérer la version formatée du texte (pour copier-coller).

    Args:
        doc_id: ID ou URL du document (doit être public ou dans le Drive LG Rez).
    """
    if len(doc_id) < 44:
        raise commons.UserInputError("doc_id", "Doit être l'ID ou l'URL d'un document Google Docs")
    elif len(doc_id) > 44:  # URL fournie (pas que l'ID)
        mtch = re.search(r"/d/([\w-]{44})(\W|$)", doc_id)
        if mtch:
            doc_id = mtch.group(1)
        else:
            raise commons.UserInputError("doc_id", "Doit être l'ID ou l'URL d'un document Google Docs")

    await journey.send("Récupération du document...")
    async with journey.channel.typing():
        content = gsheets.get_doc_content(doc_id)

    formatted_text = ""
    for (_text, style) in content:
        _text = _text.replace("\v", "\n").replace("\f", "\n")

        # Espaces/newlines au début/fin de _text ==> à part
        motif = re.fullmatch(r"(\s*)(.*?)(\s*)", _text)
        pref, text, suff = motif.group(1, 2, 3)

        if not text:  # espaces/newlines uniquement
            formatted_text += pref + suff
            continue

        # Remplacement des mentions
        for i in (3, 2, 1, 0):
            text = re.sub(rf"@([\w-]+( [\w-]+){{{i}}})", _joueur_repl, text)
        text = re.sub(r"@(\w+)", _role_repl, text)
        text = re.sub(r":(\w+):", _emoji_repl, text)

        if style.get("bold"):
            text = tools.bold(text)
        if style.get("italic"):
            text = tools.ital(text)
        if style.get("strikethrough"):
            text = tools.strike(text)
        if style.get("smallCaps"):
            text = text.upper()
        if wff := style.get("weightedFontFamily"):
            if wff["fontFamily"] in ["Consolas", "Courier New"]:
                text = tools.code(text)
        if link := style.get("link"):
            if url := link.get("url"):
                if "://" not in text:
                    text = text + f" (<{url}>)"
        elif style.get("underline"):  # ne pas souligner si lien
            text = tools.soul(text)

        formatted_text += pref + text + suff

    await tools.send_blocs(journey.channel, formatted_text)
    choice = await journey.buttons(
        f"————————————————————",
        {
            "publish": discord.ui.Button(label="Publier sur #annonces", emoji="📣"),
            "get_raw": discord.ui.Button(label="Récupérer la version formatée", emoji="📝"),
            "stop": discord.ui.Button(label="Stop", emoji="⏹"),
        },
    )

    if choice == "publish":
        await tools.send_blocs(config.Channel.annonces, formatted_text)
        await journey.send("Fait !")
    elif choice == "get_raw":
        await journey.send(formatted_text, code=True)
    else:
        await journey.send("Mission aborted.")


@app_commands.context_menu(name="Modifier ce message")
@tools.mjs_only
@journey_context_menu
async def modif(journey: DiscordJourney, message: discord.Message):
    """Modifie un message du bot (COMMANDE MJ)

    Il n'est pas possible (pour le moment ?) de modifier une image, pièce jointe ou embed.
    """
    if message.author == journey.member:
        await journey.send("Pour modifier ton propre message, utilise le petit crayon\n(blaireau)", ephemeral=True)
        return

    if message.author != config.bot.user:
        await journey.send(":x: Les messages des autres membres ne sont pas modifiables !", ephemeral=True)
        return

    (content,) = await journey.modal(
        "Modifier le message",
        discord.ui.TextInput(
            label="Contenu", style=discord.TextStyle.paragraph, default=message.content, min_length=1, max_length=2000
        ),
    )
    await message.edit(content=content)
